from typing import TypeVar, Type, Generic, Callable, Any

from amino.case import Case

from lenses import UnboundLens
from amino import Left, Right
from amino.lenses.lens import lens

from ribosome.data.plugin_state import PluginState
from ribosome.config.component import ComponentData
from ribosome.compute.tpe_data import (ProgType, UnknownProgType, StateProgType, AffiliationProgType, RootProgType,
                                       ComponentProgType, MainDataProgType, PlainMainDataProgType,
                                       InternalMainDataProgType, StateProg, ResourcesStateProgType, PlainStateProgType,
                                       RibosomeStateProgType)
from ribosome.config.resources import Resources
from ribosome.compute.wrap_data import ProgWrappers
from ribosome.compute.ribosome import Ribosome

A = TypeVar('A')
D = TypeVar('D')
M = TypeVar('M')
R = TypeVar('R')
CC = TypeVar('CC')
C = TypeVar('C')
PlS = PluginState[D, CC]  # type: ignore


class zoom_main_data(
        Generic[D, CC, M],
        Case[MainDataProgType, Callable[[PlS], M]],
        alg=MainDataProgType,
):

    def plain_main_data_prog_type(self, tpe: PlainMainDataProgType[M]) -> Callable[[PlS], D]:
        return lambda a: a.data

    def internal_main_data_prog_type(self, tpe: InternalMainDataProgType[D, CC]) -> Callable[[PlS], PlS]:
        return lambda a: a


class pick_main_data(
        Generic[D, CC, M],
        Case[MainDataProgType[M], Callable[[PlS, M], PlS]],
        alg=MainDataProgType,
):

    def plain_main_data_prog_type(self, tpe: PlainMainDataProgType[M]) -> Callable[[PlS, D], PlS]:
        return lambda ps, data: ps.set.data(data)

    def internal_main_data_prog_type(self, tpe: InternalMainDataProgType[D, CC]) -> Callable[[PlS, PlS], PlS]:
        return lambda ps, data: data


# TODO optimize
def data_for_type(ps: PluginState[D, CC], tpe: Type[C]) -> C:
    return ps.data_by_type(tpe)


class wrap_affiliation(
        Generic[C, D, CC, M],
        Case[AffiliationProgType, Callable[[PluginState[D, CC]], C]],
        alg=AffiliationProgType,
):

    def root_prog_type(self, tpe: RootProgType[M]) -> Callable[[PluginState[D, CC]], M]:
        return zoom_main_data.match(tpe.main)

    def component_prog_type(self, tpe: ComponentProgType[C, D]) -> Callable[[PlS], ComponentData[M, Any]]:
        inner: Callable[[PluginState[D, CC]], M] = zoom_main_data.match(tpe.main)
        def wrap(ps: PluginState[D, CC]) -> ComponentData[M, Any]:
            return ComponentData(inner(ps), data_for_type(ps, tpe.comp))
        return wrap


class unwrap_affiliation(
        Generic[C, D, CC, M],
        Case[AffiliationProgType[M, C], Callable[[PlS, C], PlS]],
        alg=AffiliationProgType,
):

    def root_prog_type(self, tpe: RootProgType[M]) -> Callable[[PlS, M], PlS]:
        return pick_main_data.match(tpe.main)

    def component_prog_type(self, tpe: ComponentProgType[M, C]) -> Callable[[PlS, ComponentData[M, Any]], PlS]:
        inner: Callable[[PlS, M], PlS] = pick_main_data.match(tpe.main)
        def unwrap(ps: PlS, data: ComponentData[M, Any]) -> PlS:
            return inner(ps, data.main).update_component_data(data.comp)
        return unwrap


def comp_get(ribo: Ribosome[D, CC, C]) -> C:
    return ribo.state.data_by_type(ribo.comp_type)


def comp_set(ribo: Ribosome[D, CC, C], comp: C) -> Ribosome[D, CC, C]:
    return ribo.mod.state(lambda ps: ps.update_component_data(comp))


def ribosome_comp_lens(tpe: Type[C]) -> UnboundLens[Ribosome[D, CC, C], Ribosome[D, CC, C], C, C]:
    return lens.Lens(comp_get, comp_set)


class wrap_resources(
        Generic[C, D, CC, M, R],
        Case[StateProgType[M, C, R], Callable[[PluginState[D, CC]], R]],
        alg=StateProgType,
):

    def resources_state_prog_type(self, tpe: ResourcesStateProgType[M, C, D, CC]) -> Callable[[PlS], Resources[D, CC]]:
        inner: Callable[[PluginState[D, CC]], C] = wrap_affiliation.match(tpe.affiliation)
        return lambda ps: ps.resources_with(inner(ps))

    def plain_state_prog_type(self, tpe: PlainStateProgType[M, C]) -> Callable[[PlS], C]:
        return wrap_affiliation.match(tpe.affiliation)

    def wrap_ribosome(self, tpe: RibosomeStateProgType[D, CC, C]) -> Callable[[PlS], Ribosome[D, CC, C]]:
        comp_lens = ribosome_comp_lens(tpe.comp)
        def wrap(ps: PlS) -> Ribosome[D, CC, C]:
            return Ribosome(ps, tpe.comp, comp_lens)
        return wrap


class unwrap_resources(
        Generic[C, D, CC, M, R],
        Case[StateProgType[M, C, R], Callable[[PluginState[D, CC], R], PluginState[D, CC]]],
        alg=StateProgType,
):

    def resources_state_prog_type(self, tpe: ResourcesStateProgType[M, C, D, CC]
                                  ) -> Callable[[PlS, Resources[D, CC]], PlS]:
        inner: Callable[[PlS, C], PlS] = unwrap_affiliation.match(tpe.affiliation)
        def unwrap(ps: PlS, res_wrapped: Resources[C, CC]) -> PlS:
            return inner(ps, res_wrapped.data)
        return unwrap

    def plain_state_prog_type(self, tpe: PlainStateProgType[M, C]) -> Callable[[PlS, R], PlS]:
        return unwrap_affiliation.match(tpe.affiliation)

    def unwrap_ribosome(self, tpe: RibosomeStateProgType[D, CC, C]) -> Callable[[PlS, Ribosome[D, CC, C]], PlS]:
        def wrap(ps: PlS, ribo: Ribosome[D, CC, C]) -> PlS:
            return ribo.state
        return wrap


class prog_wrappers(
        Generic[M, C, R, D, CC],
        Case[ProgType[M, C, R], ProgWrappers[PlS, R]],
        alg=ProgType,
):

    def unknown_prog_type(self, tpe: UnknownProgType[M, C]) -> ProgWrappers[PlS, R]:
        return Left('not an `NvimIOState`')

    def state_prog(self, tpe: StateProg[M, C, R, A]) -> ProgWrappers[PlS, R]:
        return Right(ProgWrappers(wrap_resources.match(tpe.tpe), unwrap_resources.match(tpe.tpe)))


__all__ = ('prog_wrappers',)



from typing import Generic, TypeVar, Callable

from amino import List, Dat, ADT, Maybe, Nil

from ribosome.rpc.data.prefix_style import PrefixStyle

A = TypeVar('A')
B = TypeVar('B')


class DocMeta(Generic[A], ADT['DocMeta[A]']):
    pass


class NoMeta(DocMeta[A]):
    pass


class Code(DocMeta[A]):

    def __init__(self, lang: str) -> None:
        self.lang = lang


class AnchorType(ADT['AnchorType']):
    pass


class VariableAnchor(AnchorType):

    def __init__(self, scope: str, prefix: bool) -> None:
        self.scope = scope
        self.prefix = prefix


class MappingAnchor(AnchorType):
    pass


class RpcAnchor(AnchorType):

    def __init__(self, prefix: PrefixStyle) -> None:
        self.prefix = prefix


class GeneralAnchor(AnchorType):
    pass


class Anchor(Dat['Anchor']):

    @staticmethod
    def cons(
            text: str,
            tpe: AnchorType,
    ) -> 'Anchor':
        return Anchor(
            text,
            tpe,
        )

    def __init__(self, text: str, tpe: AnchorType) -> None:
        self.text = text
        self.tpe = tpe


class Headline(DocMeta[A]):

    @staticmethod
    def cons(level: int=1, anchor: Anchor=None) -> 'Headline[A]':
        return Headline(level, Maybe.optional(anchor))

    def __init__(self, level: int, anchor: Maybe[Anchor]) -> None:
        self.level = level
        self.anchor = anchor


class CustomDocMeta(DocMeta[A]):

    def __init__(self, extra: A) -> None:
        self.extra = extra


class LinkType(ADT['LinkType']):
    pass


class WebLinkType(LinkType):
    pass


class DocLinkType(LinkType):
    pass


class Link(Generic[A], DocMeta[A]):

    def __init__(self, target: str, tpe: LinkType, desc: Maybe[str]) -> None:
        self.target = target
        self.tpe = tpe
        self.desc = desc


class DocFragment(Generic[A], ADT['DocFragment[A]']):
    pass


class DocCat(DocFragment[A]):

    def __init__(self, fragments: List[DocFragment]) -> None:
        self.fragments = fragments


class DocString(DocFragment[A]):

    @staticmethod
    def none(text: str) -> 'DocString[A]':
        return DocString(text, NoMeta())

    def __init__(self, text: str, meta: DocMeta[A]) -> None:
        self.text = text
        self.meta = meta


class DocLine(Generic[A], Dat['DocLine[A]']):

    @staticmethod
    def string(text: str) -> 'DocLine[A]':
        return DocLine(DocString.none(text))

    @staticmethod
    def empty() -> 'DocLine[A]':
        return DocLine.string('')

    @staticmethod
    def headline(text: str, level: int, anchor: Anchor=None) -> 'DocLine[A]':
        return DocLine(DocString(text, Headline.cons(level, anchor)))

    def __init__(self, data: DocFragment) -> None:
        self.data = data


class DocBlock(Generic[A], Dat['DocBlock[A]']):

    @staticmethod
    def none(lines: List[DocLine[A]]) -> 'DocBlock[A]':
        return DocBlock(lines, NoMeta())

    @staticmethod
    def string(text: str) -> 'DocBlock[A]':
        return DocBlock.none(List(DocLine.string(text)))

    @staticmethod
    def headline(text: str, level: int, anchor: Anchor=None) -> 'DocLine[B]':
        return DocBlock.none(List(DocLine.headline(text, level, anchor)))

    @staticmethod
    def empty() -> 'DocLine[A]':
        return DocBlock.none(List(DocLine.empty()))

    def __init__(self, lines: List[DocLine[A]], meta: DocMeta[A]) -> None:
        self.lines = lines
        self.meta = meta


class StaticDoc(Dat['StaticDoc']):

    @staticmethod
    def cons(
            pre: List[DocBlock[A]]=Nil,
            post: List[DocBlock[A]]=Nil,
    ) -> 'StaticDoc':
        return StaticDoc(
            pre,
            post,
        )

    def __init__(self, pre: List[DocBlock[A]], post: List[DocBlock[A]]) -> None:
        self.pre = pre
        self.post = post


class DocCompiler(Generic[A, B], Dat['DocCompiler[A, B]']):

    def __init__(self, compile: Callable[[List[DocLine[A]], B], List[str]], conf: B) -> None:
        self.compile = compile
        self.conf = conf


__all__ = ('DocMeta', 'NoMeta', 'Headline', 'Code', 'CustomDocMeta', 'LinkType', 'WebLinkType', 'DocLinkType', 'Link',
           'DocFragment', 'DocCat', 'DocString', 'DocLine', 'DocBlock', 'StaticDoc', 'DocCompiler',)



from typing import Callable, TypeVar, Tuple, Any, Type

from amino import Either, do, Do, _, IO, List
from amino.logging import module_log

from ribosome.compute.tpe import prog_type
from ribosome.compute.wrap import prog_wrappers
from ribosome.compute.output import (ProgOutput, ProgOutputUnit, ProgOutputResult, ProgOutputIO, ProgScalarIO,
                                     ProgGatherIOs, ProgScalarSubprocess, ProgGatherSubprocesses, Echo, ProgIOEcho,
                                     GatherSubprocesses, GatherIOs, Gather, ProgGather)
from ribosome.nvim.io.state import NS
from ribosome.compute.wrap_data import ProgWrappers
from ribosome.config.basic_config import NoData
from ribosome.compute.ribosome import Ribosome
from ribosome.compute.tpe_data import ribo_state_prog
from ribosome.compute.prog import Prog
from ribosome.compute.program import Program, ProgramBlock, ProgramCompose
from ribosome.process import Subprocess
from ribosome.rpc.args import ParamsSpec
from ribosome.rpc.api import RpcProgram
from ribosome.rpc.arg_parser import ArgParser, JsonArgParser, TokenArgParser

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
P = TypeVar('P')
R = TypeVar('R')
CC = TypeVar('CC')
C = TypeVar('C')


def arg_parser(rpc_program: RpcProgram, params_spec: ParamsSpec) -> ArgParser:
    tpe = JsonArgParser if rpc_program.options.json else TokenArgParser
    return tpe(params_spec)


def parse_args(rpc_program: RpcProgram, args: List[Any]) -> NS[D, List[Any]]:
    return arg_parser(rpc_program, rpc_program.program.params_spec).parse(args)


def prog_type_error(func: Callable[[P], NS[R, A]], error: str) -> None:
    raise Exception(f'program `{func.__name__}` has invalid type: {error}')


def program_from_data(
        func: Callable[[P], NS[R, A]],
        params_spec: ParamsSpec,
        wrappers: ProgWrappers,
        interpreter: ProgOutput[A, B],
        original_module: str=None,
        original_name: str=None,
) -> Program[A]:
    name = original_name or func.__name__
    return Program.cons(
        name,
        ProgramBlock(name, func, wrappers, interpreter),
        params_spec,
        original_module,
        original_name,
    )


def prog_state(
        func: Callable[[P], NS[R, A]],
        interpreter: ProgOutput[A, B],
        original_module: str=None,
        original_name: str=None,
) -> Program:
    params_spec = ParamsSpec.from_function(func)
    wrappers = prog_type(func, params_spec).value_or(lambda err: prog_type_error(func, err))
    return Program.cons(
        func.__name__,
        ProgramBlock(func.__name__, func, wrappers, interpreter),
        params_spec,
        original_module,
        original_name,
    )


def func_state_data(f: Callable[..., NS[R, A]]) -> Tuple[ParamsSpec, ProgWrappers]:
    params_spec = ParamsSpec.from_function(f)
    @do(Either[str, ProgWrappers])
    def wrappers() -> Do:
        comp_type = yield params_spec.state_type.to_either('no state type')
        state_prog = ribo_state_prog(comp_type)
        yield prog_wrappers.match(state_prog)
    wrappers = wrappers().value_or(lambda err: prog_type_error(f, err))
    return params_spec, wrappers


class prog_io:

    def __new__(cls, func: Callable[[P], NS[D, IO[A]]]) -> Program[A]:
        return prog_state(func, ProgOutputIO(ProgScalarIO()))

    @staticmethod
    def gather(func: Callable[[P], NS[D, GatherIOs[A]]]) -> Program[List[Either[str, A]]]:
        return prog_state(func, ProgOutputIO(ProgGatherIOs()))


class prog_subproc:

    def __new__(cls, func: Callable[[P], NS[D, Subprocess[A]]]) -> Program[A]:
        return prog_state(func, ProgOutputIO(ProgScalarSubprocess()))

    @staticmethod
    def gather(func: Callable[[P], NS[D, GatherSubprocesses[A]]]) -> Program[List[Either[str, A]]]:
        return prog_state(func, ProgOutputIO(ProgGatherSubprocesses()))


class ProgApi:

    def __call__(self, func: Callable[[P], NS[D, A]]) -> Program[A]:
        return prog_state(func, ProgOutputResult())

    def unit(self, func: Callable[[P], NS[D, None]]) -> Program[None]:
        return prog_state(func, ProgOutputUnit())

    def result(self, func: Callable[[P], NS[D, A]]) -> Program[A]:
        return prog_state(func, ProgOutputResult())

    def strict(self, func: Callable[[P], A]) -> Program[A]:
        def wrap(*p: P) -> NS[NoData, A]:
            return NS.pure(func(*p))
        return prog_state(wrap, ProgOutputResult(), func.__module__, func.__name__)

    def do(self, rettype: Type[A]=None) -> Callable[[Callable[..., Do]], Program[A]]:
        def do_wrap(func: Callable[..., Do]) -> Program[A]:
            f = do(Prog[A])(func)
            params_spec = ParamsSpec.from_function(f)
            return Program.cons(func.__name__, ProgramCompose(f), params_spec)
        return do_wrap

    io = prog_io
    subproc = prog_subproc

    def comp(self, f: Callable[[P], NS[C, A]]) -> Program[A]:
        params_spec, wrappers = func_state_data(f)
        @do(NS[Ribosome[D, CC, C], A])
        def zoomed(*p: P) -> Do:
            comp_lens = yield NS.inspect(_.comp_lens)
            yield f(*p).zoom(comp_lens)
        return program_from_data(zoomed, params_spec, wrappers, ProgOutputResult(), f.__module__, f.__name__)

    def echo(self, f: Callable[[P], NS[D, Echo]]) -> Program[None]:
        return prog_state(f, ProgOutputIO(ProgIOEcho()))

    def gather(self, func: Callable[[P], NS[D, Gather[A]]]) -> Program[List[Either[str, A]]]:
        return prog_state(func, ProgOutputIO(ProgGather()))


prog = ProgApi()


__all__ = ('prog',)



from typing import TypeVar, Any, Type, Callable

from amino import Try, Lists, Either, do, Do

from msgpack import ExtType
from amino.util.string import decode as decode_data

from ribosome.nvim.io.compute import NvimIO, NvimIOPure, NvimIORequest, NRParams
from ribosome.nvim.io.trace import NvimIOException
from ribosome.nvim.io.compute import NvimIOError
from ribosome.nvim.io.cons import nvimio_recover_fatal, nvimio_from_either

A = TypeVar('A')


def nvim_error_msg(exc: Exception) -> str:
    return Try(lambda: decode_data(exc.args[0])) | str(exc)


def nvim_request_error(name: str, args: tuple, desc: str, error: Any) -> NvimIO[A]:
    msg = (nvim_error_msg(error.cause) if isinstance(error, NvimIOException) else str(error))
    return NvimIOError(f'{desc} in nvim request `{name}({Lists.wrap(args).join_comma})`: {msg}')


@decode_data.register(ExtType)
def decode_ext_type(a: ExtType) -> ExtType:
    return a


@do(NvimIO[Either[str, Any]])
def nvim_nonfatal_request(name: str, *args: Any, params: NRParams=NRParams.cons()) -> Do:
    request: NvimIORequest[Any] = NvimIORequest(name, Lists.wrap(args), params)
    value = yield nvimio_recover_fatal(request, lambda a: nvim_request_error(name, args, 'fatal error', a))
    yield (
        nvimio_from_either(Try(value.map, decode_data).lmap(str))
        if params.decode else
        NvimIOPure(value)
    )


@do(NvimIO[A])
def nvim_request(name: str, *args: Any, params: NRParams=NRParams.cons()) -> Do:
    result = yield nvim_nonfatal_request(name, *args, params=params)
    yield nvimio_from_either(result)


def nvim_sync_request(name: str, *args: Any, params: NRParams=NRParams.cons()) -> NvimIO[A]:
    return nvim_request(name, *args, params=params.set.sync(True))


@do(NvimIO[A])
def typechecked_request(name: str, tpe: Type[A], *args: Any) -> Do:
    raw = yield nvim_sync_request(name, *args)
    yield (
        NvimIOPure(raw)
        if isinstance(raw, tpe) else
        NvimIOError(f'invalid result type of request {name}{args}: {raw}')
    )


@do(NvimIO[A])
def data_cons_request(
        name: str,
        cons: Callable[[Any], Either[str, A]],
        *args: Any,
        params: NRParams=NRParams.cons(),
) -> Do:
    raw = yield nvim_sync_request(name, *args, params=params)
    parsed = yield cons(raw)
    yield nvimio_from_either(parsed)


def data_cons_request_strict(
        name: str,
        cons: Callable[[Any], Either[str, A]],
        *args: Any,
        params: NRParams=NRParams.cons(),
) -> NvimIO[A]:
    return data_cons_request(name, lambda a: NvimIOPure(cons(a)), *args, params=params)


@do(NvimIO[Either[str, A]])
def data_cons_request_nonfatal(
        name: str,
        cons: Callable[[Either[str, Any]], Either[str, A]],
        *args: Any,
        params: NRParams=NRParams.cons(),
) -> Do:
    raw = yield nvim_nonfatal_request(name, *args, params=params.set.sync(True))
    return cons(raw)


__all__ = ('nvim_nonfatal_request', 'nvim_request', 'typechecked_request', 'data_cons_request',
           'data_cons_request_nonfatal', 'data_cons_request',)



from typing import Callable, TypeVar

from amino import Either, do, Do, Right
from amino.util.tpe import first_type_arg, type_arg, normalize_generic_type, is_subclass, is_subclass
from amino.tc.base import normalize_type

from ribosome.config.resources import Resources
from ribosome.config.component import ComponentData
from ribosome.data.plugin_state import PluginState
from ribosome.compute.tpe_data import (MainDataProgType, InternalMainDataProgType, PlainMainDataProgType,
                                       ComponentProgType, AffiliationProgType, PlainStateProgType,
                                       ResourcesStateProgType, StateProg, ProgType, UnknownProgType, RootProgType,
                                       RibosomeStateProgType)
from ribosome.compute.ribosome import Ribosome
from ribosome.compute.wrap import prog_wrappers
from ribosome.compute.wrap_data import ProgWrappers
from ribosome.nvim.io.state import NS
from ribosome.rpc.args import ParamsSpec

A = TypeVar('A')
P = TypeVar('P')
R = TypeVar('R')


def main_data_prog(data_type: type) -> MainDataProgType:
    return (
        InternalMainDataProgType()
        if is_subclass(data_type, PluginState) else
        PlainMainDataProgType()
    )


@do(Either[str, MainDataProgType])
def component_prog(affiliation_type: type) -> Do:
    tpe = yield first_type_arg(affiliation_type)
    component_data = yield type_arg(affiliation_type, 1)
    return ComponentProgType(main_data_prog(tpe), component_data)


@do(Either[str, AffiliationProgType])
def affiliation_prog(affiliation_type: type) -> Do:
    yield (
        component_prog(affiliation_type)
        if is_subclass(affiliation_type, ComponentData) else
        Right(RootProgType(main_data_prog(affiliation_type)))
    )


@do(Either[str, PlainStateProgType])
def plain_prog(plain_type: type) -> Do:
    affiliation = yield affiliation_prog(plain_type)
    return PlainStateProgType(affiliation)


@do(Either[str, ResourcesStateProgType])
def resources_prog(resources_type: type) -> Do:
    tpe = yield first_type_arg(resources_type)
    affiliation = yield affiliation_prog(tpe)
    return ResourcesStateProgType(affiliation)


@do(Either[str, RibosomeStateProgType])
def ribosome_prog(ribosome_type: type) -> Do:
    tpe = yield type_arg(ribosome_type, 2)
    return RibosomeStateProgType(tpe)


@do(Either[str, StateProg])
def state_prog(state_type: type, return_type: type) -> Do:
    state_prog_type = yield (
        resources_prog(state_type)
        if is_subclass(state_type, Resources) else
        ribosome_prog(state_type)
        if is_subclass(state_type, Ribosome) else
        plain_prog(state_type)
    )
    return StateProg(state_prog_type, return_type)


def analyse_prog_tpe(params: ParamsSpec) -> Either[str, ProgType]:
    return params.state_type.cata(
        lambda a: Right(UnknownProgType()),
        lambda a: state_prog(a, params.return_type),
    )


@do(Either[str, ProgWrappers])
def prog_type(func: Callable[[P], NS[R, A]], params_spec: ParamsSpec) -> Do:
    tpe = yield analyse_prog_tpe(params_spec)
    yield prog_wrappers.match(tpe)


__all__ = ('analyse_prog_tpe', 'prog_type')



from typing import Callable, TypeVar, Generic

from amino import Dat, List, Nil, Maybe

D = TypeVar('D')


class NoData(Dat['NoData']):
    pass


class BasicConfig(Generic[D], Dat['BasicConfig[D]']):

    @staticmethod
    def cons(
            name: str,
            prefix: str=None,
            state_ctor: Callable[[], D]=None,
            core_components: List[str]=Nil,
            default_components: List[str]=Nil,
            internal_component: bool=True,
            settings_module: str=None,
    ) -> 'BasicConfig':
        return BasicConfig(
            name,
            prefix or name,
            state_ctor or NoData,
            core_components.cons('internal') if internal_component else core_components,
            default_components,
            Maybe.optional(settings_module),
        )

    def __init__(
            self,
            name: str,
            prefix: str,
            state_ctor: Callable,
            core_components: List[str],
            default_components: List[str],
            settings_module: Maybe[str],
    ) -> None:
        self.name = name
        self.prefix = prefix
        self.state_ctor = state_ctor
        self.core_components = core_components
        self.default_components = default_components
        self.settings_module = settings_module


__all__ = ('NoData', 'BasicConfig')



from typing import TypeVar, Callable, Tuple
from queue import Queue

from amino import do, Do, IO, List, Nil, Maybe, Nothing, Just
from amino.logging import module_log
from amino.case import Case
from amino.lenses.lens import lens

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N
from ribosome.nvim.io.state import NS
from ribosome.util.menu.prompt.data import (InputState, InputResources, InputChar, PrintableChar, SpecialChar,
                                            PromptInputAction, PromptInput, PromptInterrupt, Prompt, ProcessPrompt,
                                            PromptEcho, PromptAction, PromptStateTrans, PromptQuit, PromptUnit,
                                            PromptUpdate, PromptUpdateChar, PromptConsumerInput, PromptUpdateConsumer,
                                            PromptInit, PromptUpdateInit)
from ribosome.nvim.api.command import nvim_atomic_commands
from ribosome.util.menu.prompt.input import input_loop
from ribosome.util.menu.prompt.interrupt import intercept_interrupt, stop_prompt, stop_prompt_s

log = module_log()
A = TypeVar('A')
B = TypeVar('B')


class process_char(Case[InputChar, NS[InputState[A, B], None]], alg=InputChar):

    def printable(self, a: PrintableChar) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.append1.actions(PromptUpdateChar(a)))

    def special(self, a: SpecialChar) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.append1.actions(PromptUpdateChar(a)))


def prompt_fragment(text: str, highlight: Maybe[str]) -> List[str]:
    hl = highlight.get_or_strict('None')
    return List(f'echohl {hl}', f'echon "{text}"')


@do(NS[InputState[A, B], None])
def redraw_prompt() -> Do:
    prompt, cursor = yield NS.inspect(lambda s: (s.prompt, s.cursor))
    pre, cursor_char, post = prompt.pre, prompt.post[:1], prompt.post[1:]
    cmds = List((pre, Nothing), (cursor_char, Just('RibosomePromptCaret')), (post, Nothing)).flat_map2(prompt_fragment)
    yield NS.lift(nvim_atomic_commands(cmds.cons('redraw')))


@do(NS[InputState[A, B], None])
def redraw_if_echoing() -> Do:
    state = yield NS.inspect(lambda a: a.state)
    yield redraw_prompt() if isinstance(state, PromptEcho) else NS.unit


class update_prompt_text(Case[InputChar, Prompt], alg=InputChar):

    def __init__(self, prompt: Prompt) -> None:
        self.prompt = prompt

    def printable(self, a: PrintableChar) -> Prompt:
        return self.prompt.copy(cursor=self.prompt.cursor + 1, pre=self.prompt.pre + a.char)

    def special(self, a: SpecialChar) -> Prompt:
        return self.prompt


@do(IO[List[PromptInputAction[A]]])
def dequeue(inputs: Queue) -> Do:
    first = yield IO.delay(inputs.get)
    tail = yield dequeue(inputs) if inputs.qsize() > 0 else IO.pure(Nil)
    return tail.cons(first)


class process_input(Case[PromptInputAction[A], Maybe[PromptUpdate[A]]], alg=PromptInputAction):

    def init(self, a: PromptInit[A]) -> Maybe[PromptUpdate[A]]:
        return Just(PromptUpdateInit())

    def input(self, a: PromptInput[A]) -> Maybe[PromptUpdate[A]]:
        return Just(PromptUpdateChar(a.char))

    def interrupt(self, a: PromptInterrupt[A]) -> Maybe[PromptUpdate[A]]:
        return Nothing

    def consumer(self, a: PromptConsumerInput[A]) -> Maybe[PromptUpdate[A]]:
        return Just(PromptUpdateConsumer(a.data))


class execute_prompt_action(Case[PromptAction, NS[InputResources[A, B], None]], alg=PromptAction):

    def state_trans(self, a: PromptStateTrans) -> NS[InputResources[A, B], None]:
        return NS.modify(lens.state.state.set(a.state))

    def quit(self, a: PromptQuit) -> NS[InputResources[A, B], None]:
        return stop_prompt_s()

    def unit(self, a: PromptUnit) -> NS[InputResources[A, B], None]:
        return NS.unit


class process_prompt_update_char(Case[PromptUpdate[B], NS[InputState[A, B], None]], alg=PromptUpdate):

    def init(self, a: PromptUpdateInit[B]) -> NS[InputState[A, B], None]:
        return NS.unit

    def char(self, update: PromptUpdateChar[B]) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.mod.prompt(lambda a: update_prompt_text(a)(update.char)))

    def consumer(self, a: PromptUpdateConsumer[B]) -> NS[InputState[A, B], None]:
        return NS.unit


@do(NS[InputResources[A, B], None])
def process_prompt_update(process: ProcessPrompt, action: PromptUpdate[B]) -> Do:
    yield process_prompt_update_char.match(action).zoom(lens.state)
    yield redraw_if_echoing().zoom(lens.state)
    action = yield process(action).zoom(lens.state)
    yield execute_prompt_action.match(action)


def update_prompt(process: ProcessPrompt) -> Callable[[List[PromptUpdate[B]]], NS[InputResources[A, B], None]]:
    def update_prompt(chars: List[PromptUpdate[B]]) -> NS[InputState[A, B], None]:
        return chars.traverse(lambda char: process_prompt_update(process, char), NS).replace(None)
    return update_prompt


class process_action(Case[PromptUpdate[B], NS[InputState[A, B], None]], alg=PromptUpdate):

    def init(self, a: PromptUpdateInit[B]) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.append1.actions(a))

    def input_char(self, a: PromptUpdateChar[B]) -> NS[InputState[A, B], None]:
        return process_char.match(a.char)

    def custom(self, a: PromptUpdateConsumer[B]) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.append1.actions(a))


def pop_actions(s: InputState[A, B]) -> NvimIO[Tuple[InputState[A, B], List[PromptUpdate[B]]]]:
    return N.pure((s.set.actions(Nil), s.actions))


@do(NS[InputResources[A, B], None])
def prompt_recurse(input: List[PromptUpdate[B]]) -> Do:
    stop = yield NS.inspect(lambda a: a.stop)
    yield input.traverse(process_action.match, NS).zoom(lens.state)
    current = yield NS.apply(pop_actions).zoom(lens.state)
    update = yield NS.inspect(lambda s: s.update)
    yield update(current)
    yield NS.unit if stop.is_set() else prompt_loop()


@do(NS[InputResources[A, B], None])
def prompt_loop() -> Do:
    inputs = yield NS.inspect(lambda s: s.inputs)
    input = yield NS.from_io(dequeue(inputs))
    actions = input.traverse(process_input.match, Maybe)
    yield actions.cata_strict(prompt_recurse, NS.unit)


@do(NS[InputResources[A, B], None])
def start_prompt_loop() -> Do:
    inputs = yield NS.inspect(lambda s: s.inputs)
    yield NS.from_io(IO.delay(inputs.put, PromptInit()))
    yield prompt_loop()


@do(NvimIO[A])
def prompt(process: ProcessPrompt, initial: A) -> Do:
    log.debug(f'running prompt with {initial}')
    res = InputResources.cons(
        InputState.cons(initial),
        update_prompt(process),
    )
    input_thread = yield N.fork(input_loop, res)
    result = yield intercept_interrupt(res.inputs, res.stop, res, start_prompt_loop().run_s(res))
    yield N.from_io(stop_prompt(res.inputs, res.stop))
    yield N.simple(input_thread.join)
    return result.state.data


__all__ = ('prompt',)



from typing import TypeVar, Generic, Callable

from amino import ADT, Maybe, Either
from amino.tc.monad import Monad
from amino.dat import ADTMeta
from amino.func import CallByName, call_by_name
from amino.tc.base import ImplicitsMeta, Implicits

from ribosome.compute.wrap_data import ProgWrappers
from ribosome.nvim.io.state import NS
from ribosome.compute.output import ProgOutput

A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
M = TypeVar('M')
P = TypeVar('P')
S = TypeVar('S')
R = TypeVar('R')


class ProgMeta(ADTMeta, ImplicitsMeta):

    @property
    def unit(self) -> 'Prog[None]':
        return Prog.pure(None)


class Prog(Generic[A], ADT['Prog[A]'], Implicits, implicits=True, auto=True, metaclass=ProgMeta):

    @staticmethod
    def from_maybe(fa: Maybe[A], error: CallByName) -> 'Prog[A]':
        return fa / Prog.pure | (lambda: Prog.error(error))

    @staticmethod
    def m(fa: Maybe[A], error: CallByName) -> 'Prog[A]':
        return Prog.from_maybe(fa, error)

    @staticmethod
    def from_either(fa: Either[str, A]) -> 'Prog[A]':
        return fa.cata(Prog.error, Prog.pure)

    @staticmethod
    def e(fa: Either[B, A]) -> 'Prog[A]':
        return Prog.from_either(fa.lmap(str))

    @staticmethod
    def pure(a: A) -> 'Prog[A]':
        return ProgPure(a)

    @staticmethod
    def error(error: CallByName) -> 'Prog[A]':
        return ProgError(call_by_name(error))


class ProgExec(Generic[A, B, S, R], Prog[B]):

    def __init__(
            self,
            name: str,
            code: NS[R, A],
            wrappers: ProgWrappers[S, R],
            output_type: ProgOutput,
    ) -> None:
        self.name = name
        self.code = code
        self.wrappers = wrappers
        self.output_type = output_type


class ProgBind(Generic[A, B], Prog[B]):

    def __init__(self, fa: Prog[A], f: Callable[[A], Prog[B]]) -> None:
        self.fa = fa
        self.f = f


class ProgPure(Generic[A], Prog[A]):

    def __init__(self, value: A) -> None:
        self.value = value


class ProgError(Generic[A], Prog[A]):

    def __init__(self, msg: str) -> None:
        self.msg = msg


class Monad_Prog(Monad, tpe=Prog):

    def pure(self, a: A) -> Prog[A]:
        return Prog.pure(a)

    def flat_map(self, fa: Prog[A], f: Callable[[A], Prog[B]]) -> Prog[B]:
        return ProgBind(fa, f)


__all__ = ('Prog', 'ProgBind', 'ProgPure', 'ProgPure', 'ProgError')



from typing import Any
import json

from amino import List, Lists

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.command import nvim_command


def format_json_cmd(args: List[str], data: dict) -> str:
    j = json.dumps(data)
    return f'{args.join_tokens} {j}'


def json_cmd(cmd: str, *args: str, **data: Any) -> NvimIO[str]:
    return nvim_command(cmd, format_json_cmd(Lists.wrap(args), data), verbose=True)


__all__ = ('json_cmd',)



from typing import Any, TypeVar

from amino import Either, List, Left, do, Right, curried, Do, Map, Maybe
from amino.mod import instance_from_module
from amino.logging import module_log

from ribosome.config.component import Component

log = module_log()
D = TypeVar('D')
CC = TypeVar('CC')


class ComponentResolver:

    def __init__(
            self,
            name: str,
            available_components: Map[str, Component[Any, CC]],
            core: List[str],
            default: List[str],
            requested: Maybe[List[str]],
    ) -> None:
        self.name = name
        self.available_components = available_components
        self.core = core
        self.default = default
        self.requested = requested

    def run(self) -> Either[str, List[Component]]:
        return self.components.traverse(self.create_components, Either)

    @property
    def components(self) -> List[str]:
        additional = self.requested | self.default
        components = self.core + additional
        log.debug(f'starting {self.name} with components {components}')
        return components

    def create_components(self, name: str) -> Either[str, List[Component]]:
        def report(errs: List[str]):
            msg = 'invalid {} component module "{}": {}'
            log.error(msg.format(self.name, name, errs))
        return self.resolve_name(name).leffect(report)

    def resolve_name(self, name: str) -> Either[List[str], Component]:
        auto = f'{self.name}.components.{name}'
        return (
            self.declared_component(name)
            .accum_error_f(lambda: self.component_from_exports(auto).lmap(List))
            .accum_error_f(lambda: self.component_from_exports(name).lmap(List))
            .flat_map(self.check_component(name))
        )

    def declared_component(self, name: str) -> Either[List[str], Component]:
        return (
            self.available_components
            .lift(name)
            .to_either(List(f'no auto component defined for `{name}`'))
        )

    @do(Either[str, Component])
    def component_from_exports(self, mod: str) -> Do:
        mod = yield Either.import_module(mod)
        yield instance_from_module(mod, Component)

    @curried
    def check_component(self, name: str, comp: Component) -> Either[str, Component]:
        return (
            Right(comp)
            if isinstance(comp, Component) else
            Left(List(f'invalid type for auto component: {comp}'))
        )


__all__ = ('ComponentResolver',)



from __future__ import annotations
from typing import Generic, TypeVar, Type

from lenses import UnboundLens

from amino import Dat

from ribosome.data.plugin_state import PluginState

D = TypeVar('D')
CC = TypeVar('CC')
C = TypeVar('C')


class Ribosome(Generic[D, CC, C], Dat['Ribosome[D, CC, C]']):

    def __init__(
            self,
            state: PluginState[D, CC],
            comp_type: Type[C],
            comp_lens: UnboundLens['Ribosome[D, CC, C]', 'Ribosome[D, CC, C]', C, C],
    ) -> None:
        self.state = state
        self.comp_type = comp_type
        self.comp_lens = comp_lens


__all__ = ('Ribosome',)



from typing import TypeVar, Generic

from amino.dat import Dat

from ribosome.config.component import Components

A = TypeVar('A')
D = TypeVar('D')
CC = TypeVar('CC')
G = TypeVar('G')


class Resources(Generic[D, CC], Dat['Resources[D, CC]']):

    def __init__(self, data: D, components: Components[CC]) -> None:
        self.data = data
        self.components = components


__all__ = ('Resources',)



from typing import Any, Callable
from threading import Thread

from amino import List, do, Do, Just, Lists, Map, Nil, IO, Either
from amino.json import dump_json
from amino.logging import module_log
from amino.io import IOException

from ribosome.rpc.api import RpcProgram
from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PS
from ribosome.rpc.to_plugin import run_program
from ribosome.rpc.data.rpc import RpcArgs
from ribosome.test.config import TestConfig
from ribosome.rpc.state import cons_state
from ribosome.nvim.io.api import N
from ribosome.logging import nvim_logging
from ribosome.nvim.io.compute import NvimIO
from ribosome.components.internal.update import update_components
from ribosome.nvim.io.data import NSuccess
from ribosome import NvimApi

log = module_log()


def program_runner(args: List[Any]) -> Callable[[RpcProgram], NS[PS, Any]]:
    def runner(program: RpcProgram) -> NS[PS, Any]:
        return run_program(program, RpcArgs.cons(args))
    return runner


def no_matching_program(method: str) -> NS[PS, Any]:
    return NS.lift(N.error(f'no matching program for {method}'))


@do(NS[PS, List[Any]])
def request(method: str, *args: Any, **json_args: Any) -> Do:
    progs = yield NS.inspect(lambda a: a.programs)
    matches = progs.filter(lambda a: a.rpc_name == method)
    json = yield NS.from_either(dump_json(Map(json_args)))
    json_arg = List(json) if json_args else Nil
    yield (
        no_matching_program(method)
        if matches.empty else
        matches.traverse(program_runner(Lists.wrap(args) + json_arg), NS)
    )


@do(NS[PS, Any])
def request_one(method: str, *args: Any) -> Do:
    results = yield request(method, *args)
    yield NS.m(results.head, lambda: f'empty result list for request `{method}`')


def fork_request(method: str, *args: Any, **json_args: Any) -> NS[PS, Either[IOException, Thread]]:
    def run(s: PS, v: NvimApi) -> None:
        try:
            result = request(method, *args, **json_args).run(s).run_a(v)
            if not isinstance(result, NSuccess):
                log.error(result)
        except Exception as e:
            log.error(e)
    return NS.apply(lambda s: N.delay(lambda v: (s, IO.fork(run, s, v, daemon=True).attempt)))


@do(NvimIO[PS])
def init_test_state(config: TestConfig) -> Do:
    log_handler = yield N.delay(nvim_logging)
    state = cons_state(config.config, config.io_interpreter, config.logger, log_handler=log_handler)
    yield update_components(Just(config.components)).nvim.run_s(state)


__all__ = ('program_runner', 'request', 'init_test_state', 'request_one', 'fork_request',)



from typing import TypeVar, Generic, Type, Any

from amino import ADT

from ribosome.data.plugin_state import PluginState
from ribosome.config.resources import Resources
from ribosome.config.component import ComponentData
from ribosome.compute.ribosome import Ribosome

A = TypeVar('A')
C = TypeVar('C')
CC = TypeVar('CC')
D = TypeVar('D')
M = TypeVar('M')
R = TypeVar('R')
S = TypeVar('S')


class MainDataProgType(Generic[M], ADT['MainDataProgType[M]']):
    pass


class InternalMainDataProgType(Generic[D, CC], MainDataProgType[PluginState[D, CC]]):
    pass


class PlainMainDataProgType(Generic[D], MainDataProgType[D]):
    pass


class AffiliationProgType(Generic[M, C], ADT['AffiliationProgType[M, C]']):
    pass


class RootProgType(Generic[M], AffiliationProgType[M, M]):

    def __init__(self, main: MainDataProgType[M]) -> None:
        self.main = main


class ComponentProgType(Generic[M, C], AffiliationProgType[M, ComponentData[M, C]]):

    def __init__(self, main: MainDataProgType[M], comp: Type[C]) -> None:
        self.main = main
        self.comp = comp


class StateProgType(Generic[M, C, R], ADT['StateProgType[M, C, R]']):
    pass


class ResourcesStateProgType(Generic[M, C, D, CC], StateProgType[M, C, Resources[D, CC]]):

    def __init__(self, affiliation: AffiliationProgType[M, C]) -> None:
        self.affiliation = affiliation


class PlainStateProgType(Generic[M, C], StateProgType[M, C, C]):

    def __init__(self, affiliation: AffiliationProgType[M, C]) -> None:
        self.affiliation = affiliation


class RibosomeStateProgType(Generic[D, CC, C], StateProgType[PluginState[D, CC], C, Ribosome[D, CC, C]]):

    def __init__(self, comp: Type[C]) -> None:
        self.comp = comp


class ProgType(Generic[M, C, R], ADT['ProgType[M, C, R]']):
    pass


class UnknownProgType(Generic[M, C], ProgType[M, C, None]):
    pass


class StateProg(Generic[M, C, R, A], ProgType[M, C, R]):

    def __init__(self, tpe: StateProgType[M, C, R], return_type: A) -> None:
        self.tpe = tpe
        self.return_type = return_type


trivial_state_prog = StateProg(PlainStateProgType(RootProgType(PlainMainDataProgType())), Any)


def ribo_state_prog(comp: Type[C]) -> StateProg[PluginState[D, CC], C, Ribosome[D, CC, C], Any]:
    return StateProg(RibosomeStateProgType(comp), Any)


__all__ = ('MainDataProgType', 'InternalMainDataProgType', 'PlainMainDataProgType', 'AffiliationProgType',
           'RootProgType', 'ComponentProgType', 'StateProgType', 'ResourcesStateProgType', 'PlainStateProgType',
           'ProgType', 'UnknownProgType', 'StateProg', 'RibosomeProg', 'trivial_state_prog', 'ribo_state_prog')



from typing import TypeVar, Callable, Generic, Any

from amino import ADT, List, Dat, Map

from ribosome.util.menu.data import MenuAction, MenuState, MenuLine
from ribosome.util.menu.prompt.data import InputState, PromptUpdate
from ribosome.nvim.io.state import NS

S = TypeVar('S')
ML = TypeVar('ML')
U = TypeVar('U')
MenuS = NS[InputState[MenuState[S, ML], U], MenuAction]


class AutoUpdate(Generic[U, ML], ADT['AutoUpdate[U, ML]']):
    pass


class AutoUpdateRefresh(AutoUpdate[U, ML]):

    def __init__(self, lines: List[MenuLine[ML]]) -> None:
        self.lines = lines


class AutoUpdateConsumer(AutoUpdate[U, ML]):

    def __init__(self, data: U) -> None:
        self.data = data


class AutoState(Generic[U, ML, S], Dat['AutoState[U, ML, S]']):

    def __init__(
            self,
            consumer: Callable[[PromptUpdate[AutoUpdate[U, ML]]], MenuS[Any, AutoUpdate[U, ML], ML]],
            state: S,
            mappings: Map[str, Callable[[], MenuS[S, AutoUpdate[U, ML], ML]]],
    ) -> None:
        self.consumer = consumer
        self.state = state
        self.mappings = mappings


AutoS = MenuS[AutoState[U, ML, S], ML, AutoUpdate[U, ML]]

__all__ = ('S', 'ML', 'U', 'MenuS', 'AutoUpdate', 'AutoUpdateRefresh', 'AutoUpdateConsumer', 'AutoState', 'AutoS',)



import os
import sys

sys.path[:0] = os.environ.get('RIBOSOME_PYTHONPATH', '').split(':')

from ribosome.nvim.api.data import NvimApi
from ribosome.logging import ribo_log

__all__ = ('in_vim', 'NvimApi', 'ribo_log')



from typing import Callable, TypeVar, Any

from amino import Either, List, do, Do, Right

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.util import cons_decode_str, cons_decode_str_list, cons_decode_str_list_option, cons_int
from ribosome.nvim.api.data import Buffer
from ribosome.nvim.io.api import N

A = TypeVar('A')
B = TypeVar('B')


def option(name: str, cons: Callable[[A], Either[str, B]]) -> NvimIO[B]:
    return N.read_cons_strict('nvim_get_option', cons, name)


def option_str(name: str) -> NvimIO[str]:
    return option(name, cons_decode_str)


def option_str_list(name: str) -> NvimIO[List[str]]:
    return option(name, cons_decode_str_list)


def option_int(name: str) -> NvimIO[int]:
    return option(name, cons_int)


def option_set(name: str, value: Any) -> NvimIO[None]:
    return N.write('nvim_set_option', name, value)


@do(NvimIO[None])
def option_modify(name: str, cons: Callable[[A], Either[str, B]], modify: Callable[[B], Either[str, B]]) -> Do:
    value = yield option(name, cons)
    new = yield N.from_either(modify(value))
    yield option_set(name, new)


def option_cat(name: str, add: List[str]) -> NvimIO[None]:
    return option_modify(name, cons_decode_str_list_option, lambda a: Right((a + add.map(str)).mk_string(',')))


def option_buffer_set(buffer: Buffer, name: str, value: Any) -> NvimIO[None]:
    return N.write('nvim_buf_set_option', buffer.data, name, value)


def cat_rtp(dir: str) -> NvimIO[None]:
    return option_cat('runtimepath', List(dir))


__all__ = ('option', 'option_str', 'option_str_list', 'option_set', 'option_modify', 'option_cat', 'option_buffer_set',
           'cat_rtp', 'option_int',)



from typing import Any, Tuple

from amino import List, do, Either, Do, Left
from amino.logging import module_log

from ribosome import NvimApi
from ribosome.rpc.comm import Comm, Rpc
from ribosome.rpc.to_vim import send_request, send_notification

log = module_log()


class RiboNvimApi(NvimApi):

    def __init__(self, name: str, comm: Comm) -> None:
        self.name = name
        self.comm = comm

    @do(Either[str, Tuple[NvimApi, Any]])
    def request(self, method: str, args: List[Any], sync: bool, timeout: float) -> Do:
        sender = send_request if sync else send_notification
        rpc_desc = 'request' if sync else 'notification'
        try:
            log.debug1(lambda: f'api: {rpc_desc} `{method}({args.join_tokens})`')
            a = sender(Rpc.nonblocking(method, args), timeout)
            comm, result = yield a.run(self.comm).attempt
            return self.copy(comm=comm), result
        except Exception as e:
            yield Left(f'request error: {e}')


__all__ = ('RiboNvimApi',)



import sys
import abc
from typing import Callable, Optional, Tuple, Any
from threading import Thread
from concurrent.futures import Future

import pyuv
from pyuv import (Pipe, Loop, Process, StdIO, UV_CREATE_PIPE, UV_READABLE_PIPE, UV_WRITABLE_PIPE,  # type: ignore
                  UV_PROCESS_WINDOWS_HIDE, TCP)

from amino import Dat, Try, Either, ADT, IO, List, do, Do, Maybe, Just, Nothing, Path
from amino.logging import module_log
from amino.case import Case

from ribosome.rpc.comm import RpcComm, OnMessage, OnError
from ribosome.rpc.error import RpcReadErrorUnknown, RpcProcessExit
from ribosome.rpc.start import start_plugin_sync, init_comm, cannot_execute_request
from ribosome.config.config import Config
from ribosome.rpc.nvim_api import RiboNvimApi

log = module_log()


class UvResources(ADT['UvResources']):

    @abc.abstractproperty
    def write_sink(self) -> Pipe:
        ...

    @abc.abstractproperty
    def read_source(self) -> Pipe:
        ...

    @abc.abstractproperty
    def error_source(self) -> Maybe[Pipe]:
        ...


class UvStdioResources(UvResources):

    def __init__(
            self,
            write: Pipe,
            read: Pipe,
    ) -> None:
        self.write = write
        self.read = read

    @property
    def write_sink(self) -> Pipe:
        return self.write

    @property
    def read_source(self) -> Pipe:
        return self.read

    @property
    def error_source(self) -> Maybe[Pipe]:
        return Nothing


class UvEmbedResources(UvResources):

    def __init__(
            self,
            write: Pipe,
            read: Pipe,
            error: Pipe,
            stdin: StdIO,
            stdout: StdIO,
            stderr: StdIO,
            proc: List[str],
    ) -> None:
        self.write = write
        self.read = read
        self.error = error
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.proc = proc

    @property
    def write_sink(self) -> Pipe:
        return self.write

    @property
    def read_source(self) -> Pipe:
        return self.read

    @property
    def error_source(self) -> Maybe[Pipe]:
        return Just(self.write)


class UvTcpResources(UvResources):

    def __init__(
            self,
            tcp: TCP,
            address: str,
            port: int,
    ) -> None:
        self.tcp = tcp
        self.address = address
        self.port = port

    @property
    def write_sink(self) -> Pipe:
        return self.tcp

    @property
    def read_source(self) -> Pipe:
        return self.tcp

    @property
    def error_source(self) -> Maybe[Pipe]:
        return Nothing


class UvSocketResources(UvResources):

    def __init__(
            self,
            pipe: Pipe,
            socket: Path,
    ) -> None:
        self.pipe = pipe
        self.socket = socket

    @property
    def write_sink(self) -> Pipe:
        return self.pipe

    @property
    def read_source(self) -> Pipe:
        return self.pipe

    @property
    def error_source(self) -> Maybe[Pipe]:
        return Nothing


class UvLoopThread(Dat['UvLoopThread']):

    def __init__(self, thread: Maybe[Thread]) -> None:
        self.thread = thread

    def update(self, thread: Thread) -> None:
        self.thread = Just(thread)

    def reset(self) -> None:
        self.thread = Nothing


class Uv(Dat['Uv']):

    @staticmethod
    def cons(
            loop: Loop,
            resources: UvResources,
            thread: UvLoopThread=None,
    ) -> 'Uv':
        return Uv(loop, resources, thread or UvLoopThread(Nothing))

    def __init__(
            self,
            loop: Loop,
            resources: UvResources,
            thread: UvLoopThread,
    ) -> None:
        self.loop = loop
        self.resources = resources
        self.thread = thread


# TODO this must stop the loop
def uv_write_error(handle: Pipe, error: Optional[int]) -> None:
    if error is not None:
        log.debug(f'uv write error: {pyuv.errno.strerror(error)}')


def uv_send(write: Pipe) -> Callable[[bytes], Either[str, None]]:
    def send(data: bytes) -> Either[str, None]:
        return Try(write.write, data, uv_write_error).lmap(lambda err: log.error(f'uv write failed: {err}'))
    return send


def uv_exit(loop: Loop) -> Callable[[], None]:
    def exit() -> None:
        Try(loop.stop).leffect(log.error)
    return exit


def uv_stdio(loop: Loop) -> UvResources:
    write = Pipe(loop)
    read = Pipe(loop)
    return UvStdioResources(write, read)


def uv_embed(loop: Loop, proc: List[str]) -> UvResources:
    write = Pipe(loop)
    read = Pipe(loop)
    error = Pipe(loop)
    stdin = StdIO(write, flags=UV_CREATE_PIPE + UV_READABLE_PIPE)
    stdout = StdIO(read, flags=UV_CREATE_PIPE + UV_WRITABLE_PIPE)
    stderr = StdIO(error, flags=UV_CREATE_PIPE + UV_WRITABLE_PIPE)
    return UvEmbedResources(write, read, error, stdin, stdout, stderr, proc)


def uv_tcp(loop: Loop, address: str, port: int) -> UvResources:
    return UvTcpResources(TCP(loop), address)


def uv_socket(loop: Loop, socket: Path) -> UvResources:
    return UvSocketResources(Pipe(loop), socket)


def on_exit(handle: Pipe, exit_status: int, term_signal: int) -> None:
    if exit_status != 0:
        log.warn(f'uv loop exited with status {exit_status}, signal {term_signal}')


OnRead = Callable[[Pipe, Optional[bytes], Optional[int]], None]


def processing_error(data: Optional[bytes]) -> Callable[[Exception], None]:
    def processing_error(error: Exception) -> None:
        if data:
            log.debug(f'{error}: {data}')
        log.error(f'error processing message from nvim: {error}')
    return processing_error


def read_pipe(on_data: OnMessage, on_error: OnError) -> OnRead:
    def read(handle: Pipe, data: Optional[bytes], error: Optional[int]) -> None:
        '''error -4095 is EOF, regular message when quitting
        '''
        if data is not None:
            on_data(data).attempt.leffect(processing_error(data))
        if error is not None:
            message = pyuv.errno.strerror(error)
            info = RpcProcessExit() if error == -4095 else RpcReadErrorUnknown(message)
            on_error(info).attempt.leffect(processing_error(None))
    return read


def uv_main_loop(uv: Uv) -> None:
    try:
        uv.loop.run()
    except Exception as e:
        log.error(e)


def start_processing(uv: Uv) -> Callable[[OnMessage, OnError], IO[None]]:
    @do(IO[None])
    def start(on_message: OnMessage, on_error: OnError) -> Do:
        on_read = read_pipe(on_message, on_error)
        yield connect_uv(uv, on_read)(uv.resources)
        yield IO.delay(uv.resources.read_source.start_read, on_read)
        thread = yield IO.fork(uv_main_loop, uv)
        yield IO.delay(uv.thread.update, thread)
    return start


@do(IO[None])
def stop_uv_loop(uv: Uv) -> Do:
    yield IO.delay(uv.loop.stop)
    yield uv.thread.thread.cata(lambda t: IO.delay(t.join, 3), IO.pure(None))
    yield IO.delay(uv.thread.reset)


def stop_processing(uv: Uv) -> Callable[[], IO[None]]:
    def stop() -> IO[None]:
        return stop_uv_loop(uv)
    return stop


def join_uv_loop(uv: Uv) -> IO[None]:
    return uv.thread.thread.cata(lambda t: IO.delay(t.join), IO.failed(f'no uv loop running'))


def uv_join(uv: Uv) -> Callable[[], IO[None]]:
    def join() -> IO[None]:
        return join_uv_loop(uv)
    return join


def cons_uv(cons_resources: Callable[[Loop], UvResources]) -> Tuple[Uv, RpcComm]:
    loop = Loop()
    resources = cons_resources(loop)
    uv = Uv.cons(loop, resources)
    comm = RpcComm(start_processing(uv), stop_processing(uv), uv_send(resources.write_sink), uv_join(uv),
                   uv_exit(loop))
    return uv, comm


def cons_uv_embed(proc: List[str]) -> Tuple[Uv, RpcComm]:
    return cons_uv(lambda loop: uv_embed(loop, proc))


def cons_uv_stdio() -> Tuple[Uv, RpcComm]:
    return cons_uv(lambda loop: uv_stdio(loop))


def cons_uv_tcp(address: str) -> Tuple[Uv, RpcComm]:
    return cons_uv(lambda loop: uv_tcp(loop, address))


def cons_uv_socket(socket: Path) -> Tuple[Uv, RpcComm]:
    return cons_uv(lambda loop: uv_socket(loop, socket))


class UvConnection(ADT['UvConnection']):
    pass


class UvEmbedConnection(UvConnection):

    def __init__(self, proc: Process) -> None:
        self.proc = proc


def nvim_connected(handle: Pipe, error: Any) -> None:
    if error:
        log.error(f'connecting to nvim: {error}')


class connect_uv(Case[UvResources, IO[UvConnection]], alg=UvResources):

    def __init__(self, uv: Uv, on_read: Callable[[Pipe, bytes, int], None]) -> None:
        self.uv = uv
        self.on_read = on_read

    @do(IO[UvConnection])
    def stdio(self, resources: UvStdioResources) -> Do:
        yield IO.delay(resources.read.open, sys.stdin.fileno())
        yield IO.delay(resources.write.open, sys.stdout.fileno())

    @do(IO[UvConnection])
    def embed(self, resources: UvEmbedResources) -> Do:
        proc = yield IO.delay(
            Process.spawn,
            self.uv.loop,
            args=resources.proc,
            exit_callback=on_exit,
            flags=UV_PROCESS_WINDOWS_HIDE,
            stdio=(resources.stdin, resources.stdout, resources.stderr),
        )
        yield IO.delay(resources.error.start_read, self.on_read)
        return UvEmbedConnection(proc)

    @do(IO[UvConnection])
    def tcp(self, resources: UvTcpResources) -> Do:
        yield IO.delay(resources.tcp.connect, (resources.address, resources.port), nvim_connected)
        yield IO.delay(resources.tcp.start_read, self.on_read)

    @do(IO[UvConnection])
    def socket(self, resources: UvSocketResources) -> Do:
        yield IO.delay(resources.pipe.connect, str(resources.socket), nvim_connected)
        yield IO.delay(resources.pipe.start_read, self.on_read)


class ConnectedUv(Dat['ConnectedUv']):

    def __init__(self, uv: Uv, connection: UvConnection) -> None:
        self.uv = uv
        self.connection = connection


def start_uv_plugin_sync(config: Config) -> IO[None]:
    uv, rpc_comm = cons_uv_stdio()
    return start_plugin_sync(config, rpc_comm)


embed_nvim_cmdline = List('nvim', '-n', '-u', 'NONE', '--embed')


@do(IO[RiboNvimApi])
def start_uv_embed_nvim_sync(name: str, extra: List[str]) -> Do:
    uv, rpc_comm = cons_uv(lambda loop: uv_embed(loop, embed_nvim_cmdline + extra))
    comm = yield init_comm(rpc_comm, cannot_execute_request)
    return RiboNvimApi(name, comm)


def start_uv_embed_nvim_sync_log(name: str, log: Path) -> IO[RiboNvimApi]:
    return start_uv_embed_nvim_sync(name, List(f'-V{log}'))


__all__ = ()



from typing import Callable, Any

from kallikrein import Expectation

from amino import do, IO, Do, Map
from amino.test import temp_dir

from ribosome.nvim.io.compute import NvimIO
from ribosome.test.integration.embed import setup_test_nvim_embed, TestConfig
from ribosome.nvim.api.rpc import nvim_quit
from ribosome.nvim.io.api import N
from ribosome.nvim.io.data import NResult
from ribosome.data.plugin_state import PS
from ribosome.nvim.io.state import NS
from ribosome.test.prog import init_test_state
from ribosome.test.run import run_test_io
from ribosome.test.integration.rpc import set_nvim_vars


@do(NvimIO[None])
def cleanup(result: NResult) -> Do:
    '''for some reason, the loop gets stuck if `quit` is called only once.
        '''
    yield nvim_quit()


@do(NvimIO[Expectation])
def run_test(config: TestConfig, io: Callable[..., NS[PS, Expectation]], *a: Any, **kw: Any) -> Do:
    default_vars = Map(ribosome_session_name='spec', ribosome_state_dir=str(temp_dir('persist')))
    yield set_nvim_vars(default_vars ** config.vars + ('components', config.components))
    yield config.pre()
    state = yield init_test_state(config)
    yield io(*a, **kw).run_a(state)


def external_state_test(config: TestConfig, io: Callable[..., NS[PS, Expectation]], *a: Any, **kw: Any) -> Expectation:
    @do(IO[Expectation])
    def run() -> Do:
        nvim = yield setup_test_nvim_embed(config)
        yield N.to_io_a(N.ensure(run_test(config, io, *a, **kw), cleanup), nvim.api)
    return run_test_io(run)


def external_test(config: TestConfig, io: Callable[..., NvimIO[Expectation]], *a: Any, **kw: Any) -> Expectation:
    return external_state_test(config, lambda: NS.lift(io()))


__all__ = ('external_test', 'external_state_test',)



import time
from typing import TypeVar, Callable, Any, Type, Tuple
from threading import Thread

from msgpack import ExtType

from amino import Either, IO, Maybe, List, Boolean, do, Do
from amino.func import CallByName

from ribosome.nvim.api.data import NvimApi
from ribosome.nvim.io.compute import NvimIORequest, NvimIOPure, NvimIOFatal, NvimIOError, NvimIO, NRParams
from ribosome.nvim.request import typechecked_request, data_cons_request_strict, nvim_request, data_cons_request
from ribosome.nvim.io.cons import (nvimio_delay, nvimio_recover_error, nvimio_recover_fatal, nvimio_wrap_either,
                                   nvimio_from_either, nvimio_suspend, nvimio_recover_failure, nvimio_ensure,
                                   nvimio_intercept)
from ribosome.nvim.io.data import NResult, NError, NFatal

A = TypeVar('A')
B = TypeVar('B')


class NMeta(type):

    @property
    def unit(self) -> NvimIO[None]:
        return N.pure(None)

    def pure(self, a: A) -> NvimIO[A]:
        return NvimIOPure(a)

    def wrap_either(self, f: Callable[[NvimApi], Either[B, A]]) -> NvimIO[A]:
        return nvimio_wrap_either(f)

    def from_either(self, e: Either[str, A]) -> NvimIO[A]:
        return nvimio_from_either(e)

    def e(self, e: Either[str, A]) -> NvimIO[A]:
        return N.from_either(e)

    def from_maybe(self, e: Maybe[A], error: CallByName) -> NvimIO[A]:
        return N.from_either(e.to_either(error))

    def m(self, e: Maybe[A], error: CallByName) -> NvimIO[A]:
        return N.from_maybe(e, error)

    def exception(self, exc: Exception) -> NvimIO[A]:
        return NvimIOFatal(exc)

    def failed(self, msg: str) -> NvimIO[A]:
        return N.exception(Exception(msg))

    def error(self, msg: str) -> NvimIO[A]:
        return NvimIOError(msg)

    def from_io(self, io: IO[A]) -> NvimIO[A]:
        return N.delay(lambda a: io.attempt.get_or_raise())

    def delay(self, f: Callable[..., A], *a: Any, **kw: Any) -> NvimIO[A]:
        return nvimio_delay(f, *a, **kw)

    def request(self, method: str, args: List[str], params: NRParams=NRParams.cons()) -> NvimIO[A]:
        return NvimIORequest(method, args, params)

    def simple(self, f: Callable[..., A], *a: Any, **kw: Any) -> NvimIO[A]:
        return N.delay(lambda v: f(*a, **kw))

    def suspend(self, f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> NvimIO[A]:
        return nvimio_suspend(f, *a, **kw)

    def read_tpe(self, cmd: str, tpe: Type[A], *args: Any) -> NvimIO[A]:
        return typechecked_request(cmd, tpe, *args)

    def read_cons(
            self,
            cmd: str,
            cons: Callable[[Any], NvimIO[Either[str, A]]],
            *args: Any,
            params: NRParams=NRParams.cons(),
    ) -> NvimIO[A]:
        return data_cons_request(cmd, cons, *args, params=params)

    def read_cons_strict(
            self,
            cmd: str,
            cons: Callable[[Any], Either[str, A]],
            *args: Any,
            params: NRParams=NRParams.cons(),
    ) -> NvimIO[A]:
        return data_cons_request_strict(cmd, cons, *args, params=params)

    def read_ext(self, cmd: str, *args: Any) -> NvimIO[ExtType]:
        return N.read_tpe(cmd, ExtType, *args)

    def write(self, cmd: str, *args: Any, params: NRParams=NRParams.cons(sync=False)) -> NvimIO[A]:
        return nvim_request(cmd, *args, params=params).replace(None)

    def intercept(self, fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[B]]) -> NvimIO[B]:
        return nvimio_intercept(fa, f)

    def safe(self, fa: NvimIO[A]) -> NvimIO[NResult[A]]:
        return N.intercept(fa, N.pure)

    def recover_error(self, fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
        return nvimio_recover_error(fa, f)

    def recover_fatal(self, fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
        return nvimio_recover_fatal(fa, f)

    def recover_failure(self, fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
        return nvimio_recover_failure(fa, f)

    def ignore_failure(self, fa: NvimIO[A]) -> NvimIO[None]:
        return nvimio_recover_failure(fa, lambda a: N.unit)

    def ensure(self, fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[None]]) -> NvimIO[A]:
        return nvimio_ensure(fa, f, lambda a: True)

    def ensure_error(self, fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[None]]) -> NvimIO[A]:
        return nvimio_ensure(fa, f, Boolean.is_a(NError))

    def ensure_fatal(self, fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[None]]) -> NvimIO[A]:
        return nvimio_ensure(fa, f, Boolean.is_a(NFatal))

    def ensure_failure(self, fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[None]]) -> NvimIO[A]:
        return nvimio_ensure(fa, f, Boolean.is_a((NError, NFatal)))

    @do(IO[Tuple[NvimApi, A]])
    def to_io(self, fa: NvimIO[A], api: NvimApi) -> Do:
        api1, result = yield IO.delay(fa.run, api)
        result_strict = yield IO.from_either(result.to_either)
        return api1, result_strict

    @do(IO[NvimApi])
    def to_io_s(self, fa: NvimIO[A], api: NvimApi) -> Do:
        api1, result = yield N.to_io(fa, api)
        return api1

    @do(IO[A])
    def to_io_a(self, fa: NvimIO[A], api: NvimApi) -> Do:
        api1, result = yield N.to_io(fa, api)
        return result

    def sleep(self, duration: float) -> NvimIO[None]:
        return N.delay(lambda v: time.sleep(duration))

    def fork(self, f: Callable[..., NvimIO[None]], *a: Any, daemon: bool=True, **kw: Any) -> NvimIO[Thread]:
        def fork(v: NvimApi) -> NvimIO[Thread]:
            return N.from_io(IO.fork(lambda: f(*a, **kw).unsafe(v), daemon=daemon))
        return N.suspend(fork)


class N(metaclass=NMeta):
    pass

__all__ = ('N',)



from typing import Tuple

from amino import List, I, do, Do, Path
from amino.logging import module_log

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.data import Tabpage, Window, Buffer
from ribosome.nvim.api.util import (cons_ext, cons_ext_list, cons_checked_list, cons_checked_e, extract_int_pair,
                                    cons_decode_bool)
from ribosome.nvim.api.command import nvim_command
from ribosome.nvim.io.api import N
from ribosome.nvim.api.function import nvim_call_cons_strict

log = module_log()


def current_tabpage() -> NvimIO[Tabpage]:
    return N.read_tpe('nvim_get_current_tabpage', cons_ext(Tabpage))


def current_window() -> NvimIO[Window]:
    return N.read_cons_strict('nvim_get_current_win', cons_ext(Window))


def current_buffer() -> NvimIO[Buffer]:
    return N.read_cons_strict('nvim_get_current_buf', cons_ext(Buffer))


def tabpages() -> NvimIO[List[Tabpage]]:
    return N.read_cons_strict('nvim_list_bufs', cons_ext_list(Tabpage))


def windows() -> NvimIO[List[Window]]:
    return N.read_cons_strict('nvim_list_wins', cons_ext_list(Window))


def buffers() -> NvimIO[List[Buffer]]:
    return N.read_cons_strict('nvim_list_bufs', cons_ext_list(Buffer))


def window_buffer(window: Window) -> NvimIO[Buffer]:
    return N.read_cons_strict('nvim_win_get_buf', cons_ext(Buffer), window.data)


def set_buffer_lines(
        buffer: Buffer,
        start: int,
        end: int,
        lines: List[str],
        strict_indexing: bool=False,
) -> NvimIO[None]:
    return N.write('nvim_buf_set_lines', buffer.data, start, end, strict_indexing, lines)


def set_buffer_content(buffer: Buffer, lines: List[str]) -> NvimIO[None]:
    return set_buffer_lines(buffer, 0, -1, lines, False)


def buffer_lines(
        buffer: Buffer,
        start: int,
        end: int,
        strict_indexing: bool=False,
) -> NvimIO[None]:
    return N.read_cons_strict(
        'nvim_buf_get_lines',
        cons_checked_list(str, I),
        buffer.data,
        start,
        end,
        strict_indexing,
    )


def buffer_content(buffer: Buffer) -> NvimIO[List[str]]:
    return buffer_lines(buffer, 0, -1, False)


@do(NvimIO[List[str]])
def current_buffer_content() -> Do:
    buffer = yield current_buffer()
    yield buffer_content(buffer)


def buffer_number(buffer: Buffer) -> NvimIO[int]:
    return N.read_tpe('nvim_buf_get_number', int, buffer.data)


@do(NvimIO[None])
def close_buffer(buffer: Buffer) -> Do:
    num = yield buffer_number(buffer)
    yield nvim_command('bdelete', num)


@do(NvimIO[None])
def close_current_buffer() -> Do:
    buffer = yield current_buffer()
    yield close_buffer(buffer)


def cursor(window: Window) -> NvimIO[Tuple[int, int]]:
    return N.read_cons_strict('nvim_win_get_cursor', cons_checked_e(list, extract_int_pair), window.data)


@do(NvimIO[Tuple[int, int]])
def current_cursor() -> Do:
    window = yield current_window()
    yield cursor(window)


@do(NvimIO[int])
def window_line(window: Window) -> Do:
    line, col = yield cursor(window)
    return line


def focus_window(window: Window) -> NvimIO[None]:
    return N.write('nvim_set_current_win', window.data)


def buffer_name(buffer: Buffer) -> NvimIO[str]:
    return N.read_tpe('nvim_buf_get_name', str, buffer.data)


def set_buffer_name(buffer: Buffer, name: str) -> NvimIO[str]:
    return N.write('nvim_buf_set_name', buffer.data, name)


@do(NvimIO[str])
def window_buffer_name(window: Window) -> Do:
    buffer = yield window_buffer(window)
    yield buffer_name(buffer)


@do(NvimIO[str])
def current_buffer_name() -> Do:
    buffer = yield current_buffer()
    yield buffer_name(buffer)


@do(NvimIO[bool])
def buflisted(buffer: Buffer) -> Do:
    num = yield buffer_number(buffer)
    yield nvim_call_cons_strict(cons_decode_bool, 'buflisted', num)


def set_cursor(window: Window, coords: Tuple[int, int]) -> NvimIO[None]:
    line, col = coords
    return N.write('nvim_win_set_cursor', window.data, coords)


def set_line(window: Window, line: int) -> NvimIO[None]:
    return set_cursor(window, (line, 0))


@do(NvimIO[None])
def set_local_cursor(coords: Tuple[int, int]) -> Do:
    window = yield current_window()
    yield set_cursor(window, coords)


def window_number(window: Window) -> NvimIO[int]:
    return N.read_tpe('nvim_win_get_number', int, window.data)


@do(NvimIO[int])
def current_window_number() -> Do:
    window = yield current_window()
    yield window_number(window)


def send_input(data: str) -> NvimIO[None]:
    return N.read_tpe('nvim_input', int, data)


def edit_file(file: Path) -> NvimIO[None]:
    return nvim_command('edit', str(file))


@do(NvimIO[None])
def echo(text: str) -> Do:
    yield N.write('nvim_out_write', f"{text}\n")


def window_height(window: Window) -> NvimIO[int]:
    return N.read_tpe('nvim_win_get_height', int, window.data)


@do(NvimIO[int])
def current_window_height() -> Do:
    window = yield current_window()
    yield window_height(window)


@do(NvimIO[None])
def close_window(window: Window) -> Do:
    num = yield window_number(window)
    yield nvim_command(f'{num}close!')


__all__ = ('current_tabpage', 'current_window', 'current_buffer', 'tabpages', 'windows', 'buffers', 'window_buffer',
           'set_buffer_lines', 'set_buffer_content', 'buffer_lines', 'buffer_content', 'current_buffer_content',
           'buffer_number', 'close_buffer', 'close_current_buffer', 'cursor', 'current_cursor', 'window_line',
           'focus_window', 'buffer_name', 'window_buffer_name', 'current_buffer_name', 'set_cursor', 'set_line',
           'set_local_cursor', 'current_window_number', 'send_input', 'edit_file', 'echo', 'window_number',
           'window_height', 'current_window_height', 'close_window', 'set_buffer_name',)



from typing import Callable, Tuple, Any

from amino import Map, List, Either, Left, Right, do, Do

from ribosome.nvim.api.data import NvimApi, StrictNvimApi

Handler = Callable[['StrictNvimApi', str, List[Any]], Either[str, Tuple[NvimApi, Any]]]


def api_info(vim: NvimApi, name: str, args: List[Any]) -> Either[str, Tuple[NvimApi, Any]]:
    return Right((vim, (1, {})))


def rh_write(vim: NvimApi, name: str, args: List[Any]) -> Either[str, Tuple[NvimApi, Any]]:
    return Right((vim, None))


@do(Either[str, Tuple[NvimApi, Any]])
def default_request_handler(vim: NvimApi, name: str, args: List[Any]) -> Do:
    handler = yield default_request_handlers.lift(name).to_either(f'no default request handler for {name}')
    yield handler(vim, name, args)


@do(Either[str, Tuple[NvimApi, Any]])
def pop_first(vim: NvimApi, name: str, args: List[Any]) -> Do:
    name1, args1 = yield args.uncons.to_either(f'invalid command: {name}({args})')
    yield default_request_handler(vim, name1, args1)


def rh_atomic(vim: NvimApi, name: str, args: List[Any]) -> Either[str, Tuple[NvimApi, Any]]:
    count = args.head.map(len).get_or_strict(0)
    return Right((vim, [None] * count))


default_request_handlers = Map({
    'silent': pop_first,
    'silent!': pop_first,
    'nvim_get_api_info': api_info,
    'nvim_command': rh_write,
    'nvim_out_write': rh_write,
    'nvim_call_function': pop_first,
    'nvim_call_atomic': rh_atomic,
})


@do(Either[str, Tuple[NvimApi, Any]])
def specific_handler(desc: str, handler: Handler, vim: StrictNvimApi, name: str, args: List[Any]) -> Do:
    name1, args1 = yield args.uncons.to_either(f'empty {desc} args: {name}')
    yield handler(vim, name1, args1)


@do(Either[str, Tuple[NvimApi, Any]])
def function_handler(handler: Handler, vim: StrictNvimApi, name: str, args: List[Any]) -> Do:
    yield specific_handler('function call', handler, vim, name, args)


@do(Either[str, Tuple[NvimApi, Any]])
def command_handler(handler: Handler, vim: StrictNvimApi, name: str, args: List[Any]) -> Do:
    yield specific_handler('command', handler, vim, name, args)


class StrictRequestHandler:

    def __init__(
            self,
            extra: Handler,
            function_handler: Handler,
            command_handler: Handler,
    ) -> None:
        self.extra = extra
        self.function_handler = function_handler
        self.command_handler = command_handler

    def __call__(self, vim: NvimApi, name: str, args: List[Any], sync: bool) -> Either[List[str], Tuple[NvimApi, Any]]:
        return (
            self.extra(vim, name, args)
            .lmap(List)
            .accum_error_lift(function_handler, self.function_handler, vim, name, args)
            .accum_error_lift(command_handler, self.command_handler, vim, name, args)
            .accum_error_lift(default_request_handler, vim, name, args)
        )


def no_handler(vim: NvimApi, name: str, args: List[Any]) -> Either[str, Tuple[NvimApi, Any]]:
    return Left(f'no handler for {name}')


__all__ = ('StrictRequestHandler', 'no_handler', 'Handler',)



from typing import Union

from amino import List

from ribosome.nvim.io.compute import NvimIO
from ribosome.test.klk.expectation import await_k_with
from ribosome.nvim.api.ui import current_buffer_content, buffers

from kallikrein import Expectation
from kallikrein.matchers.lines import have_lines
from kallikrein.matchers.length import have_length
from kallikrein.matcher import Matcher


def current_buffer_matches(matcher: Matcher[List[str]]) -> NvimIO[Expectation]:
    return await_k_with(matcher, current_buffer_content)


def current_buffer_contains(lines: Union[str, List[str]]) -> NvimIO[Expectation]:
    return current_buffer_matches(have_lines(lines))


def buffer_count_is(count: int) -> NvimIO[Expectation]:
    return await_k_with(have_length(count), buffers)


__all__ = ('current_buffer_contains', 'buffer_count_is', 'current_buffer_matches',)



from typing import Callable, Any, TypeVar

from kallikrein.expectation import LiftExpectationResult, ExpectationResult, FatalSpecResult
from kallikrein import Expectation, k
from kallikrein.matcher import Matcher

from amino import do, Do

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.util import nvimio_repeat_timeout
from ribosome.nvim.io.api import N
from ribosome.nvim.io.data import NError, NResult

A = TypeVar('A')


def failure_result(expectation: Callable[..., NvimIO[Expectation]], *a: Any, **kw: Any) -> NvimIO[ExpectationResult]:
    def failure_result(result: NError[Expectation]) -> NvimIO[ExpectationResult]:
        return expectation(*a, **kw)
    return failure_result


@do(NvimIO[ExpectationResult])
def eval_expectation(expectation: Callable[..., NvimIO[Expectation]], *a: Any, **kw: Any) -> Do:
    exp = yield expectation(*a, **kw)
    yield N.from_io(exp.evaluate)


def failed_expectation(result: NResult[A]) -> NvimIO[ExpectationResult]:
    return N.pure(FatalSpecResult('await_k', Exception(str(result))))


@do(NvimIO[Expectation])
def await_k(
        expectation: Callable[..., NvimIO[Expectation]],
        *a: Any,
        timeout: int=1,
        interval: float=.25,
        **kw: Any,
) -> Do:
    yield N.recover_error(
        nvimio_repeat_timeout(
            lambda: N.recover_failure(eval_expectation(expectation, *a, **kw), failed_expectation),
            lambda a: a.success,
            f'expectation not satisfied within {timeout} seconds',
            timeout,
            interval,
        ).map(LiftExpectationResult),
        failure_result(expectation, *a, **kw),
    )


@do(NvimIO[Expectation])
def await_k_with(
        matcher: Matcher[A],
        thunk: Callable[..., NvimIO[A]],
        *a: Any,
        timeout: int=1,
        interval: float=.25,
        **kw: Any,
) -> Do:
    yield await_k(lambda: thunk(*a, **kw).map(lambda a: k(a).must(matcher)), timeout=timeout, interval=interval)


__all__ = ('await_k', 'await_k_with',)



from typing import Generic, TypeVar, Callable, Tuple, cast, Type, Any

from lenses import UnboundLens

from amino.tc.base import ImplicitsMeta, Implicits
from amino.tc.monad import Monad
from amino.tc.zip import Zip
from amino.instances.list import ListTraverse
from amino import List, curried
from amino.util.string import ToStr
from amino.state.base import StateT
from ribosome.nvim.io.compute import NvimIO

import abc
from amino.tc.base import TypeClass, tc_prop
from amino.state import State, EitherState
from ribosome.nvim.api.data import NvimApi
from ribosome.nvim.io.api import N
from amino import IO, Maybe, Either
from amino.func import CallByName
E = TypeVar('E')

A = TypeVar('A')
B = TypeVar('B')
S = TypeVar('S')
R = TypeVar('R')
ST1 = TypeVar('ST1')


monad: Monad = cast(Monad, Monad.fatal(NvimIO))


class NvimIOStateCtor(Generic[S]):

    def inspect(self, f: Callable[[S], A]) -> 'NvimIOState[S, A]':
        def g(s: S) -> NvimIO[Tuple[S, A]]:
            return monad.pure((s, f(s)))
        return NvimIOState.apply(g)

    def inspect_f(self, f: Callable[[S], NvimIO[A]]) -> 'NvimIOState[S, A]':
        def g(s: S) -> NvimIO[Tuple[S, A]]:
            return f(s).map(lambda a: (s, a))
        return NvimIOState.apply(g)

    def pure(self, a: A) -> 'NvimIOState[S, A]':
        return NvimIOState.apply(lambda s: monad.pure((s, a)))

    def delay(self, fa: Callable[..., A], *a: Any, **kw: Any) -> 'NvimIOState[S, A]':
        return NvimIOState.apply(lambda s: monad.pure((s, fa(*a, **kw))))

    def lift(self, fa: NvimIO[A]) -> 'NvimIOState[S, A]':
        def g(s: S) -> NvimIO[Tuple[S, A]]:
            return fa.map(lambda a: (s, a))
        return NvimIOState.apply(g)

    def modify(self, f: Callable[[S], S]) -> 'NvimIOState[S, None]':
        return NvimIOState.apply(lambda s: monad.pure((f(s), None)))

    def modify_f(self, f: Callable[[S], NvimIO[S]]) -> 'NvimIOState[S, None]':
        return NvimIOState.apply(lambda s: f(s).map(lambda a: (a, None)))

    def get(self) -> 'NvimIOState[S, S]':
        return self.inspect(lambda a: a)

    @property
    def unit(self) -> 'NvimIOState[S, None]':
        return NvimIOState.pure(None)

    def io(self, f: Callable[[NvimApi], A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.delay(f))

    def delay(self, f: Callable[[NvimApi], A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.delay(f))

    def suspend(self, f: Callable[[NvimApi], NvimIO[A]]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.suspend(f))

    def from_io(self, io: IO[A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.wrap_either(lambda v: io.attempt))

    def from_id(self, st: State[S, A]) -> 'NvimIOState[S, A]':
        return st.transform_f(NvimIOState, lambda s: N.pure(s.value))

    def from_maybe(self, a: Maybe[B], err: CallByName) -> 'NvimIOState[S, B]':
        return NvimIOState.lift(N.from_maybe(a, err))

    m = from_maybe

    def from_either(self, e: Either[str, A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.from_either(e))

    e = from_either

    def from_either_state(self, st: EitherState[E, S, A]) -> 'NvimIOState[S, A]':
        return st.transform_f(NvimIOState, lambda s: N.from_either(s))

    def failed(self, e: str) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.failed(e))

    def error(self, e: str) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.error(e))

    def inspect_maybe(self, f: Callable[[S], Maybe[A]], err: CallByName) -> 'NvimIOState[S, A]':
        return NvimIOState.inspect_f(lambda s: N.from_maybe(f(s), err))

    def inspect_either(self, f: Callable[[S], Either[str, A]]) -> 'NvimIOState[S, A]':
        return NvimIOState.inspect_f(lambda s: N.from_either(f(s)))

    def simple(self, f: Callable[..., A], *a: Any, **kw: Any) -> 'NvimIOState[S, A]':
        return NS.lift(N.simple(f, *a, **kw))

    def sleep(self, duration: float) -> 'NvimIOState[S, None]':
        return NS.lift(N.sleep(duration))


class NvimIOStateMeta(ImplicitsMeta):

    def cons(self, run_f: NvimIO[Callable[[S], NvimIO[Tuple[S, A]]]]) -> 'NvimIOState[S, A]':
        return self(run_f)

    def apply(self, f: Callable[[S], NvimIO[Tuple[S, A]]]) -> 'NvimIOState[S, A]':
        return self.cons(monad.pure(f))

    def apply_f(self, run_f: NvimIO[Callable[[S], NvimIO[Tuple[S, A]]]]) -> 'NvimIOState[S, A]':
        return self.cons(run_f)

    def inspect(self, f: Callable[[S], A]) -> 'NvimIOState[S, A]':
        def g(s: S) -> NvimIO[Tuple[S, A]]:
            return monad.pure((s, f(s)))
        return self.apply(g)

    def inspect_f(self, f: Callable[[S], NvimIO[A]]) -> 'NvimIOState[S, A]':
        def g(s: S) -> NvimIO[Tuple[S, A]]:
            return f(s).map(lambda a: (s, a))
        return self.apply(g)

    def pure(self, a: A) -> 'NvimIOState[S, A]':
        return self.apply(lambda s: monad.pure((s, a)))

    def reset(self, s: S, a: A) -> 'NvimIOState[S, A]':
        return self.apply(lambda _: monad.pure((s, a)))

    def reset_t(self, t: Tuple[S, A]) -> 'NvimIOState[S, A]':
        return self.apply(lambda _: monad.pure(t))

    def delay(self, fa: Callable[..., A], *a: Any, **kw: Any) -> 'NvimIOState[S, A]':
        return self.apply(lambda s: monad.pure((s, fa(*a, **kw))))

    def lift(self, fa: NvimIO[A]) -> 'NvimIOState[S, A]':
        def g(s: S) -> NvimIO[Tuple[S, A]]:
            return fa.map(lambda a: (s, a))
        return self.apply(g)

    def modify(self, f: Callable[[S], S]) -> 'NvimIOState[S, None]':
        return self.apply(lambda s: monad.pure((f(s), None)))

    def modify_f(self, f: Callable[[S], NvimIO[S]]) -> 'NvimIOState[S, None]':
        return self.apply(lambda s: f(s).map(lambda a: (a, None)))

    def set(self, s: S) -> 'NvimIOState[S, None]':
        return self.modify(lambda s0: s)

    def get(self) -> 'NvimIOState[S, S]':
        return self.inspect(lambda a: a)

    @property
    def unit(self) -> 'NvimIOState[S, None]':
        return NvimIOState.pure(None)

    def s(self, tpe: Type[S]) -> NvimIOStateCtor[S]:
        return NvimIOStateCtor()

    def io(self, f: Callable[[NvimApi], A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.delay(f))

    def delay(self, f: Callable[[NvimApi], A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.delay(f))

    def suspend(self, f: Callable[[NvimApi], NvimIO[A]]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.suspend(f))

    def from_io(self, io: IO[A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.wrap_either(lambda v: io.attempt))

    def from_id(self, st: State[S, A]) -> 'NvimIOState[S, A]':
        return st.transform_f(NvimIOState, lambda s: N.pure(s.value))

    def from_maybe(self, a: Maybe[B], err: CallByName) -> 'NvimIOState[S, B]':
        return NvimIOState.lift(N.from_maybe(a, err))

    m = from_maybe

    def from_either(self, e: Either[str, A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.from_either(e))

    e = from_either

    def from_either_state(self, st: EitherState[E, S, A]) -> 'NvimIOState[S, A]':
        return st.transform_f(NvimIOState, lambda s: N.from_either(s))

    def failed(self, e: str) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.failed(e))

    def error(self, e: str) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.error(e))

    def inspect_maybe(self, f: Callable[[S], Maybe[A]], err: CallByName) -> 'NvimIOState[S, A]':
        return NvimIOState.inspect_f(lambda s: N.from_maybe(f(s), err))

    def inspect_either(self, f: Callable[[S], Either[str, A]]) -> 'NvimIOState[S, A]':
        return NvimIOState.inspect_f(lambda s: N.from_either(f(s)))

    def simple(self, f: Callable[..., A], *a: Any, **kw: Any) -> 'NvimIOState[S, A]':
        return NS.lift(N.simple(f, *a, **kw))

    def sleep(self, duration: float) -> 'NvimIOState[S, None]':
        return NS.lift(N.sleep(duration))


class NvimIOState(Generic[S, A], StateT, ToStr, Implicits, implicits=True, auto=True, metaclass=NvimIOStateMeta):

    def __init__(self, run_f: NvimIO[Callable[[S], NvimIO[Tuple[S, A]]]]) -> None:
        self.run_f = run_f

    @property
    def cls(self) -> Type['NvimIOState[S, A]']:
        return type(self)

    def run(self, s: S) -> NvimIO[Tuple[S, A]]:
        return self.run_f.flat_map(lambda f: f(s))

    def run_s(self, s: S) -> NvimIO[S]:
        return self.run(s).map(lambda a: a[0])

    def run_a(self, s: S) -> NvimIO[S]:
        return self.run(s).map(lambda a: a[1])

    def _arg_desc(self) -> List[str]:
        return List(str(self.run_f))

    def flat_map_f(self, f: Callable[[A], NvimIO[B]]) -> 'NvimIOState[S, B]':
        def h(s: S, a: A) -> NvimIO[Tuple[S, B]]:
            return f(a).map(lambda b: (s, b))
        def g(fsa: NvimIO[Tuple[S, A]]) -> NvimIO[Tuple[S, B]]:
            return fsa.flat_map2(h)
        run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
        return self.cls.apply_f(run_f1)

    def transform(self, f: Callable[[Tuple[S, A]], Tuple[S, B]]) -> 'NvimIOState[S, B]':
        def g(fsa: NvimIO[Tuple[S, A]]) -> NvimIO[Tuple[S, B]]:
            return fsa.map2(f)
        run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
        return self.cls.apply_f(run_f1)

    def transform_s(self, f: Callable[[R], S], g: Callable[[R, S], R]) -> 'NvimIOState[R, A]':
        def trans(sfsa: Callable[[S], NvimIO[Tuple[S, A]]], r: R) -> NvimIO[Tuple[R, A]]:
            s = f(r)
            return sfsa(s).map2(lambda s, a: (g(r, s), a))
        return self.cls.apply_f(self.run_f.map(curried(trans)))

    def transform_f(self, tpe: Type[ST1], f: Callable[[NvimIO[Tuple[S, A]]], Any]) -> ST1:
        def trans(s: S) -> Any:
            return f(self.run(s))
        return tpe.apply(trans)  # type: ignore

    def zoom(self, l: UnboundLens) -> 'NvimIOState[R, A]':
        return self.transform_s(l.get(), lambda r, s: l.set(s)(r))

    transform_s_lens = zoom

    def read_zoom(self, l: UnboundLens) -> 'NvimIOState[R, A]':
        return self.transform_s(l.get(), lambda r, s: r)

    transform_s_lens_read = read_zoom

    def flat_map(self, f: Callable[[A], 'NvimIOState[S, B]']) -> 'NvimIOState[S, B]':
        return Monad_NvimIOState.flat_map(self, f)


def run_function(s: NvimIOState[S, A]) -> NvimIO[Callable[[S], NvimIO[Tuple[S, A]]]]:
    try:
        return s.run_f
    except Exception as e:
        if not isinstance(s, NvimIOState):
            raise TypeError(f'flatMapped {s} into NvimIOState')
        else:
            raise


class NvimIOStateMonad(Monad, tpe=NvimIOState):

    def pure(self, a: A) -> NvimIOState[S, A]:  # type: ignore
        return NvimIOState.pure(a)

    def flat_map(  # type: ignore
            self,
            fa: NvimIOState[S, A],
            f: Callable[[A], NvimIOState[S, B]]
    ) -> NvimIOState[S, B]:
        def h(s: S, a: A) -> NvimIO[Tuple[S, B]]:
            return f(a).run(s)
        def g(fsa: NvimIO[Tuple[S, A]]) -> NvimIO[Tuple[S, B]]:
            return fsa.flat_map2(h)
        def i(sfsa: Callable[[S], NvimIO[Tuple[S, A]]]) -> Callable[[S], NvimIO[Tuple[S, B]]]:
            return lambda a: g(sfsa(a))
        run_f1 = run_function(fa).map(i)
        return NvimIOState.apply_f(run_f1)


Monad_NvimIOState = NvimIOStateMonad()


class NvimIOStateZip(Zip, tpe=NvimIOState):

    def zip(
            self,
            fa: NvimIOState[S, A],
            fb: NvimIOState[S, A],
            *fs: NvimIOState[S, A],
    ) -> NvimIOState[S, List[A]]:
        v = ListTraverse().sequence(List(fa, fb, *fs), NvimIOState)  # type: ignore
        return cast(NvimIOState[S, List[A]], v)


NS = NvimIOState


class ToNvimIOState(TypeClass):

    @abc.abstractproperty
    def nvim(self) -> NS:
        ...


class IdStateToNvimIOState(ToNvimIOState, tpe=State):

    @tc_prop
    def nvim(self, fa: State[S, A]) -> NS:
        return NvimIOState.from_id(fa)


class EitherStateToNvimIOState(ToNvimIOState, tpe=EitherState):

    @tc_prop
    def nvim(self, fa: EitherState[E, S, A]) -> NS:
        return NvimIOState.from_either_state(fa)


__all__ = ('NvimIOState',)



from typing import TypeVar, Callable, Any

from amino import Either, Boolean, do, Do
from amino.state import State
from amino.boolean import true

from ribosome.nvim.api.data import NvimApi
from ribosome.nvim.io.compute import NvimIOSuspend, NvimIOPure, NvimIOError, NvimIO, NvimIORecover, lift_n_result
from ribosome.nvim.io.data import NError, NFatal, NResult

A = TypeVar('A')
B = TypeVar('B')


def nvimio_suspend(f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> NvimIO[A]:
    return NvimIOSuspend.cons(State.inspect(lambda vim: f(vim, *a, **kw)))


def nvimio_delay(f: Callable[..., A], *a: Any, **kw: Any) -> NvimIO[A]:
    return nvimio_suspend(lambda vim: NvimIOPure(f(vim, *a, **kw)))


def nvimio_recover_error(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
    return NvimIORecover(fa, f, Boolean.is_a(NError))


def nvimio_intercept(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[B]]) -> NvimIO[B]:
    return NvimIORecover(fa, f, lambda r: true)


def nvimio_recover_fatal(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
    return NvimIORecover(fa, f, Boolean.is_a(NFatal))


def nvimio_recover_failure(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
    return NvimIORecover(fa, f, Boolean.is_a((NError, NFatal)))


def nvimio_ensure(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[None]], pred: Callable[[NResult[A]], bool]
                  ) -> NvimIO[A]:
    @do(NvimIO[A])
    def effect(error: NResult[A]) -> Do:
        yield f(error)
        yield lift_n_result.match(error)
    return NvimIORecover(fa, effect, pred)


def nvimio_wrap_either(f: Callable[[NvimApi], Either[B, A]]) -> NvimIO[A]:
    return nvimio_suspend(lambda v: f(v).cata(NvimIOError, NvimIOPure))


def nvimio_from_either(e: Either[str, A]) -> NvimIO[A]:
    return e.cata(NvimIOError, NvimIOPure)


__all__ = ('nvimio_delay', 'nvimio_wrap_either', 'nvimio_from_either', 'nvimio_suspend', 'nvimio_recover_error',
           'nvimio_recover_fatal', 'nvimio_recover_failure', 'nvimio_ensure', 'nvimio_intercept')



from typing import Callable, Any, Tuple, TypeVar
import subprocess

from kallikrein import Expectation, k
from kallikrein.matchers.lines import have_lines

from chiasma.tmux import Tmux
from chiasma.test.tmux_spec import tmux_spec_socket
from chiasma.io.compute import TmuxIO
from chiasma.commands.server import kill_server
from chiasma.test.terminal import start_tmux
from chiasma.commands.pane import send_keys, capture_pane

from amino import do, Do, IO, Lists, List, Path, env
from amino.logging import module_log
from amino.string.hues import green
from amino.test import temp_dir, fixture_path
from amino.test.path import pkg_dir
from amino.util.fs import mkdir

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N
from ribosome.nvim.api.rpc import nvim_quit
from ribosome.nvim.io.data import NResult
from ribosome.test.config import TestConfig
from ribosome.test.integration.rpc import nvim_cmdline, set_nvim_vars, setup_test_nvim_socket, test_env_vars
from ribosome.test.run import run_test_io, plugin_started
from ribosome.nvim.api.exists import wait_until_valid
from ribosome import NvimApi
from ribosome.test.integration.start import start_plugin_embed

log = module_log()
A = TypeVar('A')


@do(NvimIO[Tmux])
def test_tmux_from_vim() -> Do:
    socket = yield N.delay(lambda v: v.config.tmux_socket.get_or_strict(tmux_spec_socket))
    return Tmux.cons(socket=socket)


@do(NvimIO[A])
def test_tmux_to_nvim(fa: TmuxIO[A]) -> Do:
    tmux = yield test_tmux_from_vim()
    yield N.from_either(fa.either(tmux).lmap(str))


@do(IO[Tuple[subprocess.Popen, Tmux, Any]])
def setup_tmux(width: int, height: int) -> Do:
    proc = yield IO.delay(start_tmux, tmux_spec_socket, width, height, True)
    tmux = Tmux.cons(tmux_spec_socket)
    yield IO.sleep(.5)
    clients_tio = TmuxIO.read('list-clients -F "#{client_name}"')
    cmd = yield IO.delay(clients_tio.unsafe, tmux)
    client = yield IO.from_maybe(cmd.head, 'no clients')
    return proc, tmux, client


@do(IO[None])
def cleanup_tmux(proc: subprocess.Popen, tmux: Tmux) -> Do:
    yield IO.delay(proc.kill)
    yield IO.delay(proc.wait)
    yield IO.delay(kill_server().result, tmux)


@do(TmuxIO[None])
def send_tmux_cmdline(cmdline: str) -> Do:
    if 'VIRTUAL_ENV' in env:
        yield send_keys(0, List(f'source $VIRTUAL_ENV/bin/activate'))
    yield send_keys(0, List(cmdline))


def tmux_env_vars(socket: Path) -> List[Tuple[str, str]]:
    global_path = env['PYTHONPATH'] | ''
    project_path = pkg_dir()
    python_path = f'{project_path}:{global_path}'
    return List(
        ('NVIM_LISTEN_ADDRESS', str(socket)),
        ('PYTHONPATH', python_path),
    )

@do(IO[NvimApi])
def setup_test_nvim_tmux(config: TestConfig, socket: Path, tmux: Tmux) -> Do:
    env_args = (test_env_vars(config) + tmux_env_vars(socket)).map2(lambda k, v: f'{k}=\'{v}\'').cons('env')
    cmd = env_args + nvim_cmdline
    yield IO.delay(send_tmux_cmdline(cmd.join_tokens).run, tmux)
    yield N.to_io_a(wait_until_valid('nvim_socket', lambda n: N.pure(socket.is_socket()), 3.), None)
    yield setup_test_nvim_socket(config, socket)


def cleanup(config: TestConfig, proc: subprocess.Popen, tmux: Tmux) -> Callable[[NResult], NvimIO[None]]:
    @do(NvimIO[None])
    def cleanup(result: NResult) -> Do:
        yield N.ignore_failure(nvim_quit())
        yield N.ignore_failure(N.from_io(cleanup_tmux(proc, tmux)))
        runtime_log = yield N.from_io(IO.delay(config.log_file.read_text))
        if runtime_log:
            log.info(green('plugin output:'))
            Lists.lines(runtime_log) % log.info
            log.info('')
    return cleanup


def tmux_plugin_test(config: TestConfig, io: Callable[..., NvimIO[Expectation]], *a: Any, **kw: Any) -> Expectation:
    @do(NvimIO[None])
    def run_nvim_io() -> Do:
        yield set_nvim_vars(config.vars + ('components', config.components))
        yield config.pre()
        yield start_plugin_embed(config)
        yield plugin_started()
        yield io(*a, **kw)
    @do(IO[Expectation])
    def run() -> Do:
        socket_dir = yield IO.delay(temp_dir, 'sockets', 'nvim')
        socket = socket_dir / Lists.random_alpha()
        proc, tmux, client = yield setup_tmux(300, 120)
        nvim = yield setup_test_nvim_tmux(config, socket, tmux)
        yield N.to_io_a(N.ensure(run_nvim_io(), cleanup(config, proc, tmux)), nvim.api)
    return run_test_io(run)


@do(NvimIO[Expectation])
def store_screenshot(path: Path, current: List[str]) -> Do:
    yield N.from_io(IO.delay(path.write_text, current.join_lines))
    return k(1) == 1


@do(NvimIO[Expectation])
def check_screenshot(path: Path, current: List[str]) -> Do:
    recorded = yield N.from_io(IO.file(path))
    return k(current).must(have_lines(recorded))


@do(NvimIO[Expectation])
def screenshot(*segments: str, delay: float=0.5, record: bool=False) -> Do:
    path = fixture_path('screenshots', *segments)
    yield N.from_io(mkdir(path.parent))
    exists = yield N.from_io(IO.delay(path.exists))
    yield N.sleep(delay)
    current = yield test_tmux_to_nvim(capture_pane(0, True))
    yield store_screenshot(path, current) if not exists or record else check_screenshot(path, current)


__all__ = ('tmux_plugin_test', 'screenshot',)



from typing import Tuple, Any

from amino import _, Either, Map, Left, Right, do, Do
from amino.state import State

from ribosome.nvim.io.compute import NvimIO, NvimIOSuspend, NvimIOPure
from ribosome.nvim.io.api import N
from ribosome.nvim.api.function import nvim_call_function, nvim_call_tpe
from ribosome.nvim.api.command import nvim_command
from ribosome import NvimApi


def plugin_name() -> NvimIO[str]:
    return N.delay(_.name)


def api_info() -> NvimIO[Tuple[int, dict]]:
    def cons(data: Any) -> Either[str, Tuple[int, Map[str, Any]]]:
        return (
            Left(f'not a tuple: {data}')
            if not isinstance(data, (list, tuple)) else
            Left(f'invalid tuple size: {data}')
            if not len(data) == 2 else
            Left(f'channel is not an int: {data}')
            if not isinstance(data[0], int) else
            Left(f'metadata is not a dict: {data}')
            if not isinstance(data[1], dict) else
            Right(data).map2(lambda a, b: (a, Map(b)))
        )
    return N.read_cons_strict('nvim_get_api_info', cons)


@do(NvimIO[int])
def channel_id() -> Do:
    channel, metadata = yield api_info()
    return channel


def rpcrequest(channel: int, method: str, *args: str) -> NvimIO[Any]:
    return nvim_call_function('rpcrequest', channel, method, args)


@do(NvimIO[Any])
def rpcrequest_current(method: str, *args: str) -> Do:
    channel = yield channel_id()
    yield rpcrequest(channel, method, *args)


def nvim_quit() -> NvimIO[None]:
    return nvim_command('qall!')


def nvim_api() -> NvimIO[NvimApi]:
    return NvimIOSuspend.cons(State.get().map(NvimIOPure))


def nvim_pid() -> NvimIO[int]:
    return nvim_call_tpe(int, 'getpid')


__all__ = ('plugin_name', 'api_info', 'channel_id', 'rpcrequest', 'rpcrequest_current', 'nvim_quit', 'nvim_api',
           'nvim_pid',)



from concurrent.futures import Future, TimeoutError
from typing import Any, Callable, TypeVar, Union, Tuple

import msgpack

from amino import do, Do, IO, Either, List, ADT
from amino.state import IOState
from amino.lenses.lens import lens
from amino.logging import module_log
from amino.case import Case

from ribosome.rpc.error import RpcReadError
from ribosome.rpc.comm import Comm, RpcComm, RpcConcurrency, Rpc
from ribosome.nvim.io.data import NResult, NSuccess, NError, NFatal
from ribosome.rpc.concurrency import register_rpc
from ribosome.rpc.data.rpc import ActiveRpc
from ribosome.rpc.response import RpcResponse, RpcSyncError, RpcSyncSuccess

log = module_log()
A = TypeVar('A')
ResponseTuple = Union[Tuple[None, str], Tuple[Any, None]]


# FIXME loop is not part of comm, but uv
def rpc_error(comm: Comm) -> Callable[[RpcReadError], None]:
    def on_error(error: RpcReadError) -> IO[None]:
        return IO.pure(None)
        # IO.delay(comm.loop.stop)
    return on_error


def exclusive_increment(concurrency: RpcConcurrency) -> int:
    with concurrency.lock:
        concurrency.requests.current_id += 1
        return concurrency.requests.current_id


def exclusive_register_callback(concurrency: RpcConcurrency, id: int, rpc: Rpc) -> IO[Future]:
    result = Future()
    yield concurrency.exclusive(concurrency.requests.to_vim.update, {id: (result, rpc)})
    return result


def increment() -> IOState[RpcConcurrency, int]:
    return IOState.inspect(exclusive_increment)


def pack_error(metadata: list, payload: list) -> Callable[[Exception], IO[bytes]]:
    def pack_error(error: Exception) -> IO[bytes]:
        return IO.delay(msgpack.packb, metadata + [f'could not serialize response `{payload}`', None])
    return pack_error


@do(IOState[RpcComm, Any])
def send_rpc(metadata: list, payload: list) -> Do:
    send = yield IOState.inspect(lambda a: a.send)
    pack = IO.delay(msgpack.packb, metadata + payload).recover_with(pack_error(metadata, payload))
    payload = yield IOState.lift(pack)
    yield IOState.delay(send, payload)


def initiate_rpc(metadata: list, rpc: Rpc) -> IOState[RpcComm, Any]:
    log.debug1(f'initiate_rpc: {rpc}')
    return send_rpc(metadata, [rpc.method.encode(), rpc.args])


def wait_for_result(result: Future, timeout: float, rpc: Rpc) -> IO[Any]:
    try:
        r = result.result(timeout)
    except TimeoutError:
        return IO.failed(f'{rpc} timed out after {timeout}s')
    except Exception as e:
        log.caught_exception('waiting for request result future', e)
        return IO.failed(f'fatal error in {rpc}')
    else:
        return IO.from_either(r.lmap(lambda a: f'{rpc} failed: {a}'))


@do(IOState[Comm, Either[str, Any]])
def send_request(rpc: Rpc, timeout: float) -> Do:
    id = yield increment().zoom(lens.concurrency)
    active_rpc = ActiveRpc(rpc, id)
    result = yield IOState.inspect_f(lambda a: register_rpc(a, a.requests.to_vim, active_rpc)).zoom(lens.concurrency)
    yield initiate_rpc([0, id], rpc).zoom(lens.rpc)
    yield IOState.lift(wait_for_result(result, timeout, rpc))


@do(IOState[Comm, Any])
def send_notification(rpc: Rpc, timeout: float) -> Do:
    yield initiate_rpc([2], rpc).zoom(lens.rpc)


class handle_response(Case[RpcResponse, IOState[Comm, None]], alg=RpcResponse):

    @do(IOState[Comm, None])
    def error(self, response: RpcSyncError) -> Do:
        payload = [response.error, None]
        yield send_rpc([1, response.request_id], payload).zoom(lens.rpc)

    @do(IOState[Comm, None])
    def success(self, response: RpcSyncSuccess) -> Do:
        payload = [None, response.result]
        yield send_rpc([1, response.request_id], payload).zoom(lens.rpc)

    def case_default(self, response: RpcResponse) -> IOState[Comm, None]:
        return IOState.unit


__all__ = ('send_request', 'send_notification', 'handle_response',)



from typing import Any, Callable

from amino import IO, List, ADT
from amino.case import Case
from amino.logging import module_log

from ribosome.rpc.data.rpc_type import RpcType, BlockingRpc, NonblockingRpc
from ribosome.rpc.data.rpc import Rpc
from ribosome.nvim.io.data import NFatal, NResult, NSuccess, NError
from ribosome import ribo_log

log = module_log()


class RpcResponse(ADT['RpcResponse']):
    pass


class RpcSyncError(RpcResponse):

    def __init__(self, error: str, request_id: int) -> None:
        self.error = error
        self.request_id = request_id


class RpcAsyncError(RpcResponse):

    def __init__(self, error: str) -> None:
        self.error = error


class RpcSyncSuccess(RpcResponse):

    def __init__(self, result: Any, request_id: int) -> None:
        self.result = result
        self.request_id = request_id


class RpcAsyncSuccess(RpcResponse):
    pass


def ensure_single_result(rpc: Rpc, tpe: BlockingRpc) -> Callable[[Any, List[Any]], RpcResponse]:
    def ensure_single_result(head: Any, tail: List[Any]) -> RpcResponse:
        return (
            RpcSyncSuccess(head, tpe.id)
            if tail.empty else
            RpcSyncError(f'multiple results for {rpc}', tpe.id)
        )
    return ensure_single_result


def no_result(rpc: Rpc, tpe: BlockingRpc) -> RpcResponse:
    return RpcSyncError(f'no result for {rpc}', tpe.id)


class validate_success_response(Case[RpcType, RpcResponse], alg=RpcType):

    def __init__(self, result: NSuccess[List[Any]], rpc: Rpc) -> None:
        self.result = result
        self.rpc = rpc

    def blocking(self, tpe: BlockingRpc) -> RpcResponse:
        return self.result.value.uncons.map2(ensure_single_result(self.rpc, tpe)).get_or(no_result, self.rpc, tpe)

    def nonblocking(self, tpe: NonblockingRpc) -> RpcResponse:
        return RpcAsyncSuccess()


class error_response(Case[RpcType, RpcResponse], alg=RpcType):

    def __init__(self, error: str) -> None:
        self.error = error

    def blocking(self, tpe: BlockingRpc) -> RpcResponse:
        return RpcSyncError(self.error, tpe.id)

    def nonblocking(self, tpe: NonblockingRpc) -> RpcResponse:
        return RpcAsyncError(self.error)


class validate_rpc_result(Case[NResult[List[Any]], RpcResponse], alg=NResult):

    def __init__(self, rpc: Rpc) -> None:
        self.rpc = rpc

    def success(self, result: NSuccess[List[Any]]) -> RpcResponse:
        return validate_success_response(result, self.rpc)(self.rpc.tpe)

    def error(self, result: NError[List[Any]]) -> RpcResponse:
        return error_response(result.error)(self.rpc.tpe)

    def fatal(self, result: NFatal[List[Any]]) -> RpcResponse:
        log.caught_exception_error(f'executing {self.rpc} from vim', result.exception)
        return error_response(f'fatal error in {self.rpc}')(self.rpc.tpe)


class report_error(Case[RpcResponse, IO[None]], alg=RpcResponse):

    def sync_error(self, response: RpcSyncError) -> IO[None]:
        return IO.delay(log.debug, response.error)

    def async_error(self, response: RpcAsyncError) -> IO[None]:
        return IO.delay(ribo_log.error, response.error)

    def case_default(self, response: RpcResponse) -> IO[None]:
        return IO.pure(None)


__all__ = ('RpcResponse', 'RpcSyncError', 'RpcAsyncError', 'RpcSyncSuccess', 'RpcAsyncSuccess', 'validate_rpc_result',
           'report_error',)



import abc
from typing import Tuple

from amino import Dat, Map, Boolean, ADT, List, Maybe, Lists
from amino.boolean import false

from ribosome.compute.program import Program


class MapMode(ADT['MapMode']):

    @abc.abstractproperty
    def mnemonic(self) -> str:
        ...


class mapmode:

    class Normal(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'n'

    class Operator(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'o'

    class Visual(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'x'

    class Select(MapMode):

        @property
        def mnemonic(self) -> str:
            return 's'

    class VisualSelect(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'v'

    class Insert(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'i'

    class CommandLine(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'c'

    class Terminal(MapMode):

        @property
        def mnemonic(self) -> str:
            return 't'

    class Language(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'l'


class Mapping(Dat['Mapping']):

    @staticmethod
    def cons(ident: str, keys: str, buffer: Boolean=false, modes: List[MapMode]=None) -> 'Mapping':
        modes1 = List(mapmode.Normal()) if modes is None else modes
        return Mapping(ident, keys, Boolean(buffer), modes1)

    def __init__(self, ident: str, keys: str, buffer: Boolean, modes: List[MapMode]) -> None:
        self.ident = ident
        self.keys = keys
        self.buffer = buffer
        self.modes = modes


class Mappings(Dat['Mappings']):

    @staticmethod
    def cons(*mappings: Tuple[Mapping, Program]) -> 'Mappings':
        return Mappings(Map(Lists.wrap(mappings).map2(lambda m, p: (m.ident, (m, p)))))

    def __init__(self, mappings: Map[str, Tuple[Mapping, Program]]) -> None:
        self.mappings = mappings

    def lift(self, mapping: Mapping) -> Maybe[Program]:
        return self.mappings.lift(mapping.ident)


__all__ = ('Mapping', 'Mappings')



from typing import TypeVar, Generic, Callable, Any, Type

from amino import ADT, Dat, List, Lists, Nil, Either, Maybe
from amino.case import Case
from amino.json.encoder import Encoder
from amino.json.data import JsonError, Json
from amino.json.encoders import encode_instance
from amino.json.decoder import Decoder
from amino.json.decoders import decode_instance

from ribosome.compute.wrap_data import ProgWrappers
from ribosome.compute.output import ProgOutput, ProgOutputResult
from ribosome.nvim.io.state import NS
from ribosome.compute.prog import Prog, ProgBind, ProgExec
from ribosome.rpc.args import ParamsSpec

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
R = TypeVar('R')


class ProgramCode(Generic[A], ADT['ProgramCode[A]']):
    pass


# TODO restrict the function to a single value parameter of type ProgramArgs[A, B], where A is a custom data type
# containing the positional args and B an optional json decoded parameter.
class ProgramBlock(Generic[A, B, D], ProgramCode[B]):

    def __init__(
            self,
            name: str,
            code: Callable[..., NS[D, A]],
            wrappers: ProgWrappers,
            interpreter: ProgOutput[A, B],
    ) -> None:
        self.name = name
        self.code = code
        self.wrappers = wrappers
        self.interpreter = interpreter


class ProgramCompose(Generic[A], ProgramCode[A]):

    def __init__(self, code: Callable[..., ProgBind[Any, A]]) -> None:
        self.code = code


class ProgramMetadata(Dat['ProgramMetadata']):

    @staticmethod
    def cons(
            name: str,
            params_spec: ParamsSpec,
            original_module: str=None,
            original_name: str=None,
    ) -> 'ProgramMetadata':
        return ProgramMetadata(name, params_spec, Maybe.optional(original_module), Maybe.optional(original_name))

    def __init__(
            self,
            name: str,
            params_spec: ParamsSpec,
            original_module: Maybe[str],
            original_name: Maybe[str],
    ) -> None:
        self.name = name
        self.params_spec = params_spec
        self.original_module = original_module
        self.original_name = original_name


class Program(Generic[A], Dat['Program[A]']):

    @staticmethod
    def lift(f: Callable[..., NS[D, A]]) -> 'Program[A]':
        return Program.cons(
            f.__name__,
            ProgramBlock(f.__name__, f, ProgWrappers.id(), ProgOutputResult()),
            ParamsSpec.from_function(f),
        )

    @staticmethod
    def cons(
            name: str,
            code: ProgramCode[A],
            params_spec: ParamsSpec,
            original_module: str=None,
            original_name: str=None,
    ) -> 'Program[A]':
        return Program(code, ProgramMetadata.cons(name, params_spec, original_module, original_name))

    def __init__(
            self,
            code: ProgramCode[A],
            metadata: ProgramMetadata,
    ) -> None:
        self.code = code
        self.metadata = metadata

    def __call__(self, *args: Any) -> Prog[A]:
        return bind_program(self, Lists.wrap(args))

    @property
    def name(self) -> str:
        return self.metadata.name

    @property
    def params_spec(self) -> ParamsSpec:
        return self.metadata.params_spec


class bind_program_code(Generic[A, B, R], Case[ProgramCode[A], Prog[B]], alg=ProgramCode):

    def __init__(self, program: Program[A], args: List[Any]) -> None:
        self.args = args

    def program_compose(self, code: ProgramCompose[A]) -> Prog[A]:
        return code.code(*self.args)

    def program_block(self, code: ProgramBlock[Any, A, R]) -> Prog[A]:
        return ProgExec(code.name, code.code(*self.args), code.wrappers, code.interpreter)


def bind_program(program: Program[A], args: List[Any]) -> Prog[B]:
    return bind_program_code(program, args)(program.code)


def bind_nullary_program(program: Program[A]) -> Prog[B]:
    return bind_program(program, Nil)


def bind_programs(programs: List[Program[A]], args: List[Any]) -> Prog[B]:
    return programs.traverse(lambda a: bind_program(a, args), Prog)


def program_module(prog: Program) -> str:
    return prog.metadata.original_module | prog.code.code.__module__


def program_name(prog: Program) -> str:
    return prog.metadata.original_name | prog.metadata.name


class ProgramEncoder(Encoder[Program], tpe=Program):

    def encode(self, prog: Program) -> Either[JsonError, Json]:
        return encode_instance(prog, Program, program_module(prog), [program_name(prog)])


class ProgramDecoder(Decoder[Program], tpe=Program):

    def decode(self, tpe: Type[Program], data: Json) -> Either[JsonError, Program]:
        return decode_instance(data, 'Program')


__all__ = ('Program', 'bind_program', 'bind_nullary_program', 'bind_programs')



from amino import do, Do, List, Maybe, Dat, Just, Nothing
from amino.logging import module_log

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.ui import (current_tabpage, current_window, window_buffer, set_buffer_content,
                                  current_window_height, set_buffer_name)
from ribosome.nvim.api.data import Window, Buffer, Tabpage
from ribosome.nvim.api.option import option_buffer_set
from ribosome.nvim.api.command import nvim_command
from ribosome.nvim.io.api import N

log = module_log()


class ScratchUi(Dat['ScratchUi']):

    def __init__(self, window: Window, previous: Window, tab: Maybe[Tabpage]) -> None:
        self.window = window
        self.tab = tab
        self.previous = previous


class ScratchBuffer(Dat['ScratchBuffer']):

    def __init__(self, buffer: Buffer, ui: ScratchUi) -> None:
        self.buffer = buffer
        self.ui = ui


@do(NvimIO[Tabpage])
def create_scratch_tab() -> Do:
    yield nvim_command('tabnew')
    yield current_tabpage()


@do(NvimIO[Window])
def create_scratch_window(vertical: bool, size: Maybe[int]) -> Do:
    cmd = 'vnew' if vertical else 'new'
    size_prefix = size.cata(str, '')
    yield nvim_command(f'{size_prefix}{cmd}')
    yield current_window()


@do(NvimIO[ScratchUi])
def create_scratch_ui(use_tab: bool, vertical: bool, size: Maybe[int]) -> Do:
    current = yield current_window()
    tab = yield create_scratch_tab() / Just if use_tab else N.pure(Nothing)
    window = yield current_window() if use_tab else create_scratch_window(vertical, size)
    return ScratchUi(window, current, tab)


@do(NvimIO[None])
def configure_scratch_buffer(buffer: Buffer, name: str) -> Do:
    yield option_buffer_set(buffer, 'buftype', 'nofile')
    yield option_buffer_set(buffer, 'bufhidden', 'wipe')
    yield option_buffer_set(buffer, 'buflisted', False)
    yield option_buffer_set(buffer, 'swapfile', False)
    yield option_buffer_set(buffer, 'modifiable', False)
    yield set_buffer_name(buffer, name)


@do(NvimIO[Buffer])
def setup_scratch_buffer(ui: ScratchUi, name: str) -> Do:
    buffer = yield window_buffer(ui.window)
    yield configure_scratch_buffer(buffer, name)
    return buffer


class CreateScratchBufferOptions(Dat['CreateScratchBufferOptions']):

    @staticmethod
    def cons(
            tab: bool=None,
            vertical: bool=None,
            size: int=None,
            wrap: bool=None,
            name: str='',
    ) -> 'CreateScratchBufferOptions':
        return CreateScratchBufferOptions(
            Maybe.optional(tab),
            Maybe.optional(vertical),
            Maybe.optional(size),
            Maybe.optional(wrap),
            name,
        )

    def __init__(
            self,
            tab: Maybe[bool],
            vertical: Maybe[bool],
            size: Maybe[int],
            wrap: Maybe[bool],
            name: str,
    ) -> None:
        self.tab = tab
        self.vertical = vertical
        self.size = size
        self.wrap = wrap
        self.name = name


@do(NvimIO[ScratchBuffer])
def create_scratch_buffer(options: CreateScratchBufferOptions) -> Do:
    ui = yield create_scratch_ui(options.tab.get_or_strict(False), options.vertical.get_or_strict(False), options.size)
    buffer = yield setup_scratch_buffer(ui, options.name)
    return ScratchBuffer(buffer, ui)


@do(NvimIO[None])
def set_scratch_buffer_content(scratch: ScratchBuffer, lines: List[str]) -> Do:
    yield option_buffer_set(scratch.buffer, 'modifiable', True)
    yield set_buffer_content(scratch.buffer, lines)
    yield option_buffer_set(scratch.buffer, 'modifiable', False)


@do(NvimIO[ScratchBuffer])
def show_in_scratch_buffer(lines: List[str], options: CreateScratchBufferOptions) -> Do:
    scratch = yield create_scratch_buffer(options)
    yield set_scratch_buffer_content(scratch, lines)
    return scratch


@do(NvimIO[int])
def scratch_buffer_height(line_count: int, max_height: float) -> Do:
    current_height = yield current_window_height()
    return min(line_count, int(max_height * current_height))


@do(NvimIO[ScratchBuffer])
def show_in_scratch_buffer_default(lines: List[str], max_height: Maybe[float]) -> Do:
    size = yield max_height.map(lambda a: scratch_buffer_height(len(lines), a)).get_or_strict(None)
    yield show_in_scratch_buffer(lines, CreateScratchBufferOptions.cons(size=size))


__all__ = ('setup_scratch_buffer', 'CreateScratchBufferOptions', 'create_scratch_buffer', 'show_in_scratch_buffer',
           'show_in_scratch_buffer_default')



from typing import TypeVar, Any, Generic

from amino import do, Do, IO, Right, Either, List, Nil, Maybe, Try, Left
from amino.logging import module_log
from amino.case import Case

from ribosome.nvim.api.function import nvim_call_cons, nvim_call_tpe
from ribosome.nvim.io.compute import NvimIO, NRParams
from ribosome.nvim.io.api import N
from ribosome.util.menu.codes import modifier_codes, special_codes
from ribosome.util.menu.prompt.data import (Input, NoInput, InputResources, InterruptInput, NormalInput, SpecialChar,
                                            PrintableChar, PromptInput)
from ribosome.util.menu.prompt.interrupt import intercept_interrupt, stop_prompt

log = module_log()
A = TypeVar('A')
B = TypeVar('B')


@do(NvimIO[List[str]])
def input_modifiers() -> Do:
    code = yield nvim_call_tpe(int, 'getcharmod')
    return (
        Nil
        if code == 0 else
        modifier_codes.filter2(lambda c, n: code % c == 0).map2(lambda c, n: n)
    )


def parse_special(key: Any) -> Maybe[str]:
    return special_codes.lift(key).map(lambda a: f'<{a}>')


def parse_printable(key: Any) -> Maybe[str]:
    return Try(chr, key).or_else_call(Try, ord, key).to_maybe


def parse_key(key: Any) -> Maybe[Either[str, str]]:
    return (
        parse_special(key).map(Left)
        .or_else_call(lambda: parse_printable(key).map(Right))
    )


@do(NvimIO[Either[str, Input]])
def analyse_key(key: Either[str, str]) -> Do:
    log.debug(f'received prompt input `{key.value}`')
    modifiers = yield input_modifiers()
    return Right(NormalInput(key.cata(lambda a: SpecialChar(a, modifiers), PrintableChar)))


def parse_nonnull_input(data: Any) -> NvimIO[Either[str, Input]]:
    return parse_key(data).map(analyse_key).get_or(Left, f'could not parse input `{data}`')


def parse_input(data: Any) -> NvimIO[Either[str, Input]]:
    return (
        N.pure(Right(NoInput()))
        if data == 0 else
        parse_nonnull_input(data)
    )


def getchar() -> NvimIO[Input]:
    return nvim_call_cons(parse_input, 'getchar', False, params=NRParams.cons(decode=False, timeout=120))


class process_input(Generic[A, B], Case[Input, NvimIO[None]], alg=Input):

    def __init__(self, res: InputResources[A, B]) -> None:
        self.res = res

    @do(NvimIO[None])
    def no(self, a: NoInput) -> Do:
        yield N.sleep(self.res.interval)
        yield input_loop_unsafe(self.res)

    @do(NvimIO[None])
    def normal(self, a: NormalInput) -> Do:
        yield N.from_io(IO.delay(self.res.inputs.put, PromptInput(a.char)))
        yield input_loop_unsafe(self.res)

    def interrupt(self, a: InterruptInput) -> NvimIO[None]:
        return stop_prompt(self.res.inputs, self.res.stop)


@do(NvimIO[None])
def input_step(res: InputResources[A, B]) -> Do:
    input = yield getchar()
    yield process_input(res)(input)


def input_loop_unsafe(res: InputResources[A, B]) -> NvimIO[None]:
    return N.unit if res.stop.is_set() else input_step(res)


def input_loop(res: InputResources[A, B]) -> NvimIO[None]:
    return intercept_interrupt(res.inputs, res.stop, None, input_loop_unsafe(res))


__all__ = ('input_loop',)



from queue import Queue
from typing import Any

import msgpack

from amino import Dat, Map, IO

from ribosome.rpc.uv.uv import OnMessage, processing_error
from ribosome.rpc.comm import OnError


class StrictRpc(Dat['StrictRpc']):

    @staticmethod
    def cons(responses: Map[str, Any]) -> 'StrictRpc':
        return StrictRpc(responses, Queue(), False)

    def __init__(self, responses: Map[str, Any], queue: Queue, running: bool) -> None:
        self.responses = responses
        self.queue = queue
        self.running = running

    def send(self, data: bytes) -> None:
        self.queue.put(msgpack.unpackb(data))

    def start_processing(self, on_message: OnMessage, on_error: OnError) -> IO[None]:
        self.running = True
        return IO.fork(self.loop, on_message, on_error)

    def loop(self, on_message: OnMessage, on_error: OnError) -> None:
        while self.running:
            try:
                a = self.queue.get(timeout=.1)
            except Exception as e:
                pass
            else:
                if len(a) == 4 and a[0] == 0:
                    response = self.responses.lift(a[2].decode()).get_or_else(None)
                    on_message(msgpack.packb([1, a[1], None, response])).attempt.lmap(processing_error)

    def stop(self) -> None:
        self.running = False


__all__ = ('StrictRpc',)



from typing import TypeVar

from amino.tc.base import tc_prop
from amino import do, Do

from chiasma.io.compute import TmuxIO
from chiasma.tmux import Tmux
from chiasma.io.state import TmuxIOState

from ribosome.nvim.io.state import ToNvimIOState, NS
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.variable import variable_prefixed_str
from ribosome.nvim.io.api import N

A = TypeVar('A')
S = TypeVar('S')


@do(NvimIO[Tmux])
def tmux_from_vim() -> Do:
    socket = yield variable_prefixed_str('tmux_socket')
    return Tmux.cons(socket=socket | None)


@do(NvimIO[A])
def tmux_to_nvim(tm: TmuxIO[A]) -> Do:
    tmux = yield tmux_from_vim()
    yield N.from_either(tm.either(tmux))


class TmuxStateToNvimIOState(ToNvimIOState, tpe=TmuxIOState):

    @tc_prop
    def nvim(self, fa: TmuxIOState[S, A]) -> NS:
        return fa.transform_f(NS, tmux_to_nvim)


__all__ = ('TmuxStateToNvimIOState',)



from typing import Any

from amino import List, Try, do, Do, Either, ADT, Left, Right, Lists
from amino.logging import module_log

log = module_log()


class Receive(ADT['Receive']):
    pass


class ReceiveRequest(Receive):

    def __init__(self, id: int, method: str, args: List[Any]) -> None:
        self.id = id
        self.method = method
        self.args = args


class ReceiveNotification(Receive):

    def __init__(self, method: str, args: list) -> None:
        self.method = method
        self.args = args


class ReceiveResponse(Receive):

    def __init__(self, id: int, data: Any) -> None:
        self.id = id
        self.data = data


class ReceiveError(Receive):

    def __init__(self, id: int, error: str) -> None:
        self.id = id
        self.error = error


class ReceiveExit(Receive):
    pass


class ReceiveUnknown(Receive):

    def __init__(self, data: Any, reason: str) -> None:
        self.data = data
        self.reason = reason


@do(Either[str, Receive])
def receive_error(id: int, payload: Any) -> Do:
    data = yield Right(payload) if isinstance(payload, list) else Left(f'error payload not a list: {payload}')
    error = yield Lists.wrap(data).lift(1).to_either_f(lambda: f'too few elements for error payload: {data}')
    return ReceiveError(id, error)


@do(Either[str, Receive])
def receive_request(data: List[Any]) -> Do:
    el1, el2, el3 = yield data.lift_all(0, 1, 2).to_either_f(lambda: f'wrong number of elements: {data.length}')
    id = yield Right(el1) if isinstance(el1, int) else Left(f'id is not an int: {el1}')
    method = yield Try(el2.decode) if isinstance(el2, bytes) else Left(f'method is not a str: {el2}')
    args = yield Right(Lists.wrap(el3)) if isinstance(el3, list) else Left(f'args is not a list: {el3}')
    return ReceiveRequest(id, method, args)


@do(Either[str, Receive])
def receive_response(data: List[Any]) -> Do:
    el1, el2, el3 = yield data.lift_all(0, 1, 2).to_either_f(lambda: f'wrong number of elements: {data.length}')
    id = yield Right(el1) if isinstance(el1, int) else Left(f'id is not an int: {el1}')
    yield (
        receive_error(id, el2)
        if el2 is not None else
        Right(ReceiveResponse(id, el3))
    )


@do(Either[str, Receive])
def receive_notification(data: List[Any]) -> Do:
    el1, el2 = yield data.lift_all(0, 1).to_either_f(lambda: f'wrong number of elements: {data.length}')
    method = yield Try(el1.decode) if isinstance(el1, bytes) else Left(f'method is not a str: {el1}')
    args = yield Right(Lists.wrap(el2)) if isinstance(el2, list) else Left(f'args is not a list: {el2}')
    return ReceiveNotification(method, args)


@do(Either[str, Receive])
def validate_receive_data(raw: Any) -> Do:
    yield Right(None) if isinstance(raw, list) else Left('not a list')
    payload = Lists.wrap(raw)
    rpc_type, data = yield payload.uncons.to_either_f(lambda: f'response is empty list')
    yield (
        receive_request(data)
        if rpc_type == 0 else
        receive_response(data)
        if rpc_type == 1 else
        receive_notification(data)
        if rpc_type == 2 else
        Left(f'invalid rpc type: {rpc_type}')
    )


def cons_receive(data: Any) -> Receive:
    return validate_receive_data(data).value_or(lambda reason: ReceiveUnknown(data, reason))


__all__ = ('Receive', 'ReceiveRequest', 'ReceiveNotification', 'ReceiveResponse', 'ReceiveError', 'ReceiveExit',
           'ReceiveUnknown', 'cons_receive')



from typing import Type, Callable, Any
from types import ModuleType

from amino import Either, _, L, amino_log, __, Path, do, Do

from amino.either import ImportFailure
from amino.logging import amino_root_file_logging, module_log
from amino.mod import instance_from_module
from amino.json import decode_json
from amino.util.exception import format_exception

from ribosome.config.config import Config
from ribosome.rpc.uv.uv import start_uv_plugin_sync

log = module_log()


def report_runtime_error(result: Any) -> int:
    amino_log.error(f'error in plugin execution: {result}')
    return 1


def run_loop_uv(config: Config) -> int:
    amino_log.debug(f'starting plugin from {config.basic}')
    result = start_uv_plugin_sync(config).attempt
    return result.cata(report_runtime_error, lambda a: 0)


def config_from_module(mod: ModuleType) -> Either[str, Config]:
    return instance_from_module(mod, Config)


log_initialized = False


def error(msg: str) -> int:
    try:
        amino_log.error(msg)
    except Exception as e:
        pass
    return 1


def import_error(e: ImportFailure, desc: str) -> int:
    return error(e.expand.join_lines)


def exception(e: Exception, desc: str) -> int:
    f = __.caught_exception_error(f'starting host from {desc}', e)
    try:
        f(amino_log)
    except Exception as e:
        pass
    return 1


def start_module_config(mod: ModuleType) -> int:
    return config_from_module(mod).cata(error, run_loop_uv)


def setup_log() -> None:
    amino_root_file_logging()


def start_from(source: str, importer: Callable[[str], Either[ImportFailure, ModuleType]], desc: str) -> int:
    try:
        setup_log()
        amino_log.debug(f'start_{desc}: {source}')
        return importer(source).cata(L(import_error)(_, source), start_module_config)
    except Exception as e:
        return exception(e, source)


def start_module(mod: str) -> int:
    return start_from(mod, Either.import_module, 'module')


def start_path(path: str) -> int:
    try:
        setup_log()
        amino_log.debug(f'start_path: {path}')
        return Either.import_path(path).cata(L(import_error)(_, path), run_loop_uv)
    except Exception as e:
        return exception(e, path)


def start_file(path: str) -> int:
    p = Path(path)
    file = p / '__init__.py' if p.is_dir() else p
    return start_from(str(file), Either.import_file, 'file')


def start_json_config(data: str) -> int:
    @do(Either[str, int])
    def decode_and_run() -> Do:
        config = yield decode_json(data)
        return run_loop_uv(config)
    try:
        setup_log()
        amino_log.debug('starting plugin from json')
        return decode_and_run().value_or(error)
    except Exception as e:
        error(str(e))
        try:
            error(format_exception(e).join_lines)
        except Exception as e:
            error(str(e))


__all__ = ('start_module', 'start_file', 'start_json_config', 'start_path',)



import abc
from typing import Any, Tuple

from amino import List, Either, Right, Map
from amino.json.decoder import decode_json_type
from amino.logging import module_log

from ribosome.rpc.args import ParamsSpec

log = module_log()


def starts_with_brace(data: Any) -> bool:
    return isinstance(data, str) and data.startswith('{')


class ArgParser(abc.ABC):

    def __init__(self, params_spec: ParamsSpec) -> None:
        self.params_spec = params_spec

    @abc.abstractmethod
    def parse(self, args: List[Any]) -> Either[str, List[Any]]:
        ...


class TokenArgParser(ArgParser):

    def parse(self, args: List[Any]) -> Either[str, List[Any]]:
        return Right(args)


class JsonArgParser(ArgParser):

    def parse(self, args: List[Any]) -> Either[str, List[Any]]:
        def pick_json(start: int) -> Tuple[str, str]:
            return args[:start], args[start:].join_tokens
        strict, json_args = args.index_where(starts_with_brace).cata(pick_json, (args, '{}'))
        tpe = self.params_spec.types.last | (lambda: Map)
        return decode_json_type(json_args, tpe) / strict.cat


__all__ = ('ArgParser', 'TokenArgParser', 'JsonArgParser')



from typing import TypeVar, Generic, Any
import logging

from amino import ADT, IO, List, Dat, Nil
from amino.dat import DatMeta
from ribosome.process import Subprocess, SubprocessResult

A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
P = TypeVar('P')
PIO = TypeVar('PIO')


class ProgOutput(Generic[A, B], ADT['ProgOutput[A, B]']):
    pass


class ProgOutputUnit(ProgOutput[None, None]):
    pass


class ProgOutputResult(Generic[A], ProgOutput[A, A]):
    pass


class ProgIO(Generic[PIO], ADT['ProgIO[PIO]']):
    pass


class ProgScalarIO(ProgIO[IO]):
    pass


class ProgGatherIOs(ProgIO[List[IO]]):
    pass


class ProgScalarSubprocess(ProgIO[Subprocess]):
    pass


class ProgGatherSubprocesses(ProgIO[List[Subprocess]]):
    pass


class GatherItem(Generic[A], ADT['GatherItem[A]']):
    pass


class GatherIO(GatherItem[A]):

    def __init__(self, io: IO[A]) -> None:
        self.io = io


class GatherSubprocess(GatherItem[A]):

    def __init__(self, subprocess: Subprocess[A]) -> None:
        self.subprocess = subprocess


class GatherResult(Generic[A], ADT['GatherResult[A]']):
    pass


class GatherIOResult(Generic[A], GatherResult[A]):

    def __init__(self, result: A) -> None:
        self.result = result


class GatherSubprocessResult(Generic[A], GatherResult[A]):

    def __init__(self, result: SubprocessResult[A]) -> None:
        self.result = result


class ProgGather(ProgIO[List[GatherItem]]):
    pass


class EchoMeta(DatMeta):

    @property
    def unit(self) -> 'Echo':
        return Echo(Nil, 0)


class Echo(Dat['Echo'], metaclass=EchoMeta):
    info_level = logging.INFO
    error_level = logging.ERROR
    debug_level = logging.DEBUG

    @staticmethod
    def info(msg: str) -> 'Echo':
        return Echo(List(msg), Echo.info_level)

    @staticmethod
    def error(msg: str) -> 'Echo':
        return Echo(List(msg), Echo.error_level)

    @staticmethod
    def debug(msg: str) -> 'Echo':
        return Echo(List(msg), Echo.debug_level)

    @staticmethod
    def debugs(msg: List[str]) -> 'Echo':
        return Echo(msg, Echo.debug_level)

    def __init__(self, messages: List[str], level: int) -> None:
        self.messages = messages
        self.level = level


class ProgIOEcho(ProgIO[None]):
    pass


class ProgIOCustom(ProgIO[Any]):
    pass


class ProgOutputIO(Generic[PIO, A], ProgOutput[PIO, A]):

    def __init__(self, io: ProgIO[PIO]) -> None:
        self.io = io


class GatherIOs(Generic[A], Dat['GatherIOs[A]']):

    def __init__(self, ios: List[IO[A]], timeout: float) -> None:
        self.ios = ios
        self.timeout = timeout


class GatherSubprocesses(Generic[A], Dat['GatherSubprocess[A]']):

    def __init__(self, subprocs: List[Subprocess[A]], timeout: float) -> None:
        self.subprocs = subprocs
        self.timeout = timeout


class Gather(Generic[A], Dat['Gather[A]']):

    def __init__(self, items: List[GatherItem[A]], timeout: float) -> None:
        self.items = items
        self.timeout = timeout


__all__ = ('ProgOutput', 'ProgOutputUnit', 'ProgOutputResult', 'ProgOutputIO', 'ProgResult', 'ProgReturn',
           'ProgIO', 'ProgScalarIO', 'ProgGatherIOs', 'ProgScalarSubprocess', 'ProgGatherSubprocesses',
           'ProgIOCustom', 'ProgOutputIO', 'GatherIOs', 'GatherSubprocesses', 'GatherItem', 'GatherIO',
           'GatherSubprocess', 'ProgGather', 'Gather', 'GatherResult', 'GatherIOResult', 'GatherSubprocessResult',)



from typing import Any

from kallikrein import Expectation, k

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N
from ribosome.nvim.api.variable import var_becomes
from ribosome.test.klk.matchers.nresult import nsuccess


def var_must_become(name: str, value: Any, timeout: float=3, interval: float=.2) -> NvimIO[Expectation]:
    return N.intercept(var_becomes(name, value, timeout, interval), lambda r: N.pure(k(r).must(nsuccess(True))))


__all__ = ('var_must_become',)
